#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **Decidedly Object-Oriented Runtime-checking (DOOR) getters** (i.e.,
low-level callables introspecting metadata pertaining to high-level
:class:`beartype.door.TypeHint` wrappers).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.door._cls.doorsuper import TypeHint
from beartype.door._cls.pep.doorpep484604 import UnionTypeHint
from beartype.door._cls.pep.doorpep586 import LiteralTypeHint
from beartype.door._cls.pep.doorpep593 import AnnotatedTypeHint
from beartype.door._cls.pep.pep484.doorpep484any import AnyTypeHint
from beartype.door._cls.pep.pep484.doorpep484class import ClassTypeHint
from beartype.door._cls.pep.pep484.doorpep484newtype import NewTypeTypeHint
from beartype.door._cls.pep.pep484.doorpep484typevar import TypeVarTypeHint
from beartype.door._cls.pep.pep484585.doorpep484585callable import (
    CallableTypeHint)
from beartype.door._cls.pep.pep484585.doorpep484585generic import (
    GenericTypeHint)
from beartype.door._cls.pep.pep484585.doorpep484585subscripted import (
    SubscriptedTypeHint)
from beartype.door._cls.pep.pep484585.doorpep484585tuple import (
    TupleFixedTypeHint,
    TupleVariableTypeHint,
)
from beartype.roar import (
    BeartypeDoorNonpepException,
    # BeartypeDoorPepUnsupportedException,
)
from beartype.typing import (
    Dict,
    Type,
)
from beartype._data.typing.datatypingport import Hint
from beartype._data.hint.sign.datahintsigncls import HintSign
from beartype._data.hint.sign.datahintsigns import (
    HintSignAnnotated,
    HintSignAny,
    HintSignCallable,
    HintSignLiteral,
    HintSignNewType,
    HintSignPep484585GenericSubbed,
    HintSignPep484585GenericUnsubbed,
    HintSignTuple,
    HintSignPep484585TupleFixed,
    HintSignTypeVar,
)
from beartype._data.hint.sign.datahintsignset import (
    HINT_SIGNS_UNSUBSCRIPTABLE)
from beartype._util.hint.pep.utilpepget import get_hint_pep_args
from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none
from beartype._util.hint.pep.utilpeptest import is_hint_pep_typing

# ....................{ GETTERS                            }....................
def get_typehint_subclass(hint: Hint) -> Type[TypeHint]:
    '''
    Concrete :class:`TypeHint` subclass handling the passed low-level unwrapped
    PEP-compliant type hint if any *or* raise an exception otherwise.

    Parameters
    ----------
    hint : Hint
        Low-level type hint to be inspected.

    Returns
    -------
    Type[TypeHint]
        Concrete subclass of the abstract :mod:`TypeHint` superclass handling
        this hint.

    Raises
    ------
    beartype.roar.BeartypeDoorNonpepException
        If this API does *not* currently support the passed hint.
    beartype.roar.BeartypeDecorHintPepSignException
        If the passed hint is *not* actually a PEP-compliant type hint.
    '''

    # ..................{ SUBCLASS                           }..................
    # Sign uniquely identifying this hint if any *OR* "None" otherwise (i.e., if
    # this hint is a PEP-noncompliant class).
    hint_sign = get_hint_pep_sign_or_none(hint)

    #FIXME: [SPEED] As a negligible optimization, globalize the
    #_HINT_SIGN_TO_TYPEHINT_CLS.get() method to avoid repeated lookups here.
    # Private concrete subclass of this ABC handling this hint if any *OR*
    # "None" otherwise (i.e., if no such subclass has been authored yet).
    wrapper_subclass = _HINT_SIGN_TO_TYPEHINT_CLS.get(hint_sign)  # type: ignore[arg-type]
    # print(f'Mapping hint {hint} sign {hint_sign} to subclass {wrapper_subclass}...')

    # If this hint appears to be currently unsupported...
    if wrapper_subclass is None:
        # If either...
        if (
            # This hint is a PEP-noncompliant isinstanceable class *OR*...
            isinstance(hint, type) or

            #FIXME: This condition is kinda intense. Should we really be
            #conflating typing attributes that aren't types with objects that
            #are types? Let's investigate exactly which kinds of type hints
            #require this and contemplate something considerably more elegant.
            # An unsupported kind of PEP-compliant type hint (e.g.,
            # "typing.TypedDict" instance)...
            is_hint_pep_typing(hint)
        # Return the concrete "TypeHint" subclass handling all such classes.
        ):
            wrapper_subclass = ClassTypeHint
            # print(f'[type fallback] hint: {repr(hint)}; sign: {repr(hint_sign)}; wrapper: {repr(wrapper_subclass)}')
        # Else, raise an exception.
        else:
            raise BeartypeDoorNonpepException(
                f'Type hint {repr(hint)} '
                f'currently unsupported by "beartype.door.TypeHint".'
            )
    # Else, this hint is supported.

    #FIXME: Instead of reducing to the inappropriate "ClassTypeHint" subclass
    #here, we should instead:
    #* Define a new "UnsubscriptedTypeHint" subclass wherever we currently
    #  define the existing "SubscriptedTypeHint" subclass.
    #* Reduce to "UnsubscriptedTypeHint" rather than "ClassTypeHint" below.

    # If it is *NOT* the case that either...
    elif not (
        # This hint is unsubscriptable (i.e., permissable as a valid hint even
        # when unsubscripted by child hints) and thus *NOT* safely reducible to
        # the "ClassTypeHint" subclass even when unsubscripted *OR*...
        hint_sign in HINT_SIGNS_UNSUBSCRIPTABLE or
        # This hint is subscripted by one or more child hints.
        get_hint_pep_args(hint)
    ):
        # Replace this inappropriate "SubscriptedTypeHint" wrapper with the more
        # appropriate "ClassTypeHint" subclass wrapping unsubscripted types.
        wrapper_subclass = ClassTypeHint
    # # In any case, this hint is supported by this concrete subclass.

    #FIXME: Alternately, it might be preferable to refactor this to resemble:
    #    if (
    #       not get_hint_pep_args(hint) and
    #       get_hint_pep_origin_type_or_none(hint) is not None
    #    ):
    #        wrapper_subclass = ClassTypeHint
    #
    #That's possibly simpler and cleaner, as it seamlessly conveys the exact
    #condition we're going for -- assuming it works, of course. *sigh*
    #FIXME: While sensible, the above approach induces non-trivial test
    #failures. Let's investigate this further at a later time, please.

    # Return this subclass.
    # print(f'Mapped hint {hint} sign {hint_sign} to subclass {wrapper_subclass}!')
    return wrapper_subclass

# ....................{ PRIVATE ~ globals                  }....................
# Further initialized below by the _init() function.
_HINT_SIGN_TO_TYPEHINT_CLS: Dict[HintSign, Type[TypeHint]] = {
    HintSignAnnotated:  AnnotatedTypeHint,
    HintSignAny:        AnyTypeHint,
    HintSignCallable:   CallableTypeHint,
    HintSignLiteral:    LiteralTypeHint,
    HintSignNewType:    NewTypeTypeHint,
    HintSignTuple:      TupleVariableTypeHint,
    HintSignPep484585TupleFixed: TupleFixedTypeHint,
    HintSignTypeVar:    TypeVarTypeHint,
    HintSignPep484585GenericSubbed:   GenericTypeHint,
    HintSignPep484585GenericUnsubbed: GenericTypeHint,
}
'''
Dictionary mapping from each sign uniquely identifying PEP-compliant type hints
to the :class:`.TypeHint` subclass handling those hints.
'''

# ....................{ PRIVATE ~ initializers             }....................
def _init() -> None:
    '''
    Initialize this submodule.
    '''

    # Isolate function-specific imports.
    from beartype._data.hint.sign.datahintsignmap import (
        HINT_SIGN_ORIGIN_ISINSTANCEABLE_TO_ARGS_LEN_RANGE)
    from beartype._data.hint.sign.datahintsignset import HINT_SIGNS_UNION

    # Fully initialize the "_HINT_SIGN_TO_TYPEHINT_CLS" global dictionary.
    #
    # For each sign in the dictionary mapping from signs uniquely identifying
    # type hint factories originating from isinstanceable types to the fixed
    # number of child type hints subscripting those factories...
    for hint_sign in HINT_SIGN_ORIGIN_ISINSTANCEABLE_TO_ARGS_LEN_RANGE.keys():
        # If this sign has *NOT* already been mapped to an existing "TypeHint"
        # subclass, map this sign to the generic private
        # "SubscriptedTypeHint" subclass.
        if hint_sign not in _HINT_SIGN_TO_TYPEHINT_CLS:
            _HINT_SIGN_TO_TYPEHINT_CLS[hint_sign] = (
                SubscriptedTypeHint)
        # Else, this sign has already been mapped to an existing "TypeHint"
        # subclass. Preserve this mapping as is.

    # For each sign uniquely identifying a union, map this sign to the
    # union-specific "TypeHint" subclass.
    for hint_sign in HINT_SIGNS_UNION:
        _HINT_SIGN_TO_TYPEHINT_CLS[hint_sign] = UnionTypeHint

    # For each concrete "TypeHint" subclass registered with this dictionary
    # (*AFTER* initializing this dictionary)...
    for typehint_cls in _HINT_SIGN_TO_TYPEHINT_CLS.values():
        # If the unqualified basename of this subclass is prefixed by an
        # underscore, this subclass is private rather than public. In this case,
        # silently ignore this private subclass and continue to the next.
        if typehint_cls.__name__.startswith('_'):
            continue
        # Else, this subclass is public.

        # Sanitize the fully-qualified module name of this public subclass from
        # the private submodule declaring this subclass (e.g.,
        # "beartype.door._cls.pep.doorpep484604.UnionTypeHint") to the public
        # "beartype.door" subpackage to both improve the readability of
        # exceptions and discourage users from violating privacy encapsulation.
        typehint_cls.__module__ = 'beartype.door'

# ....................{ MAIN                               }....................
# Initialize this submodule.
_init()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`557`-compliant **class-specific utility function** unit
tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.cls.pep.clspep557` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ raiser                     }....................
def test_die_unless_type_pep557_dataclass() -> None:
    '''
    Test the
    :func:`beartype._util.cls.pep.clspep557.die_unless_type_pep557_dataclass`
    raiser.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep557Exception
    from beartype._util.cls.pep.clspep557 import (
        die_unless_type_pep557_dataclass)
    from beartype_test.a00_unit.data.pep.data_pep557 import DataclassDefault
    from pytest import raises

    # ....................{ PASS                           }....................
    # Implicitly assert that this raiser raises *NO* exception when passed a
    # dataclass type.
    die_unless_type_pep557_dataclass(DataclassDefault) is True

    # ....................{ FAIL                           }....................
    # Assert that this raiser raises the expected exception when passed a
    # non-dataclass type.
    with raises(BeartypeDecorHintPep557Exception):
        die_unless_type_pep557_dataclass(int)

    # Assert that this raiser raises the expected exception when passed a
    # non-type.
    with raises(BeartypeDecorHintPep557Exception):
        die_unless_type_pep557_dataclass(
            'Spaces of fire, and all the yawn of hell.—')

# ....................{ TESTS ~ tester                     }....................
def test_is_type_pep557_dataclass() -> None:
    '''
    Test the
    :func:`beartype._util.cls.pep.clspep557.is_type_pep557_dataclass` tester.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar._roarexc import BeartypeDecorHintPep557Exception
    from beartype._util.cls.pep.clspep557 import is_type_pep557_dataclass
    from beartype_test.a00_unit.data.pep.data_pep557 import DataclassDefault
    from pytest import raises

    # ....................{ PASS                           }....................
    # Assert that this tester returns true when passed a dataclass type.
    assert is_type_pep557_dataclass(DataclassDefault) is True

    # Assert that this tester returns false when passed a non-dataclass type.
    assert is_type_pep557_dataclass(str) is False

    # ....................{ FAIL                           }....................
    # Assert that this tester raises the expected exception when passed a
    # non-type.
    with raises(BeartypeDecorHintPep557Exception):
        is_type_pep557_dataclass('The wilderness has a mysterious tongue')


def test_is_pep557_dataclass_frozen() -> None:
    '''
    Test the
    :func:`beartype._util.cls.pep.clspep557.is_pep557_dataclass_frozen` tester.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar._roarexc import BeartypeDecorHintPep557Exception
    from beartype._util.cls.pep.clspep557 import is_pep557_dataclass_frozen
    from beartype_test.a00_unit.data.pep.data_pep557 import (
        DataclassDefault,
        DataclassFrozen,
    )
    from pytest import raises

    # ....................{ PASS                           }....................
    # Assert that this tester returns false when passed a default dataclass.
    assert is_pep557_dataclass_frozen(DataclassDefault) is False

    # Assert that this tester returns true when passed a frozen dataclass.
    assert is_pep557_dataclass_frozen(DataclassFrozen) is True

    # ....................{ FAIL                           }....................
    # Assert that this tester raises the expected exception when passed a
    # non-dataclass.
    with raises(BeartypeDecorHintPep557Exception):
        is_pep557_dataclass_frozen(str)

# ....................{ TESTS ~ getter                     }....................
def test_get_pep557_dataclass_kwargs() -> None:
    '''
    Test the
    :func:`beartype._util.cls.pep.clspep557.get_pep557_dataclass_kwargs` getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar._roarexc import BeartypeDecorHintPep557Exception
    from beartype._util.cls.pep.clspep557 import get_pep557_dataclass_kwargs
    from beartype_test.a00_unit.data.pep.data_pep557 import (
        DataclassDefault,
        DataclassFrozen,
    )
    from pytest import raises

    # ....................{ CALLABLES                      }....................
    def _assert_dataclass_kwargs_default(datacls_kwargs: object) -> None:
        '''
        Assert that the passed object returned by the
        :func:`.get_pep557_dataclass_kwargs` getter encapsulates all keyword
        parameters configuring the default dataclass *except* the ``frozen``
        parameter, which the caller is required to manually assert.
    
        Note that the public API of the standard :mod:`dataclasses` module is
        extremely deficient and currently provides no sane means of
        differentiating default from non-default keyword parameters. Our only
        recourse is to manually test the set of all instance variables known to
        be defined by the object returned by this getter under Python >= 3.9.0.
        '''

        assert dataclass_default_kwargs.init is True
        assert dataclass_default_kwargs.repr is True
        assert dataclass_default_kwargs.eq is True
        assert dataclass_default_kwargs.order is False
        assert dataclass_default_kwargs.unsafe_hash is False

    # ....................{ PASS                           }....................
    # Assert that this getter returns an object encapsulating all keyword
    # parameters configuring the default dataclass.
    dataclass_default_kwargs = get_pep557_dataclass_kwargs(DataclassDefault)
    assert dataclass_default_kwargs.frozen is False
    _assert_dataclass_kwargs_default(dataclass_default_kwargs)

    # Assert that this getter returns an object encapsulating all keyword
    # parameters configuring a frozen dataclass.
    dataclass_frozen_kwargs = get_pep557_dataclass_kwargs(DataclassFrozen)
    assert dataclass_frozen_kwargs.frozen is True
    _assert_dataclass_kwargs_default(dataclass_frozen_kwargs)

    # ....................{ FAIL                           }....................
    # Assert that this getter raises the expected exception when passed a
    # non-dataclass.
    with raises(BeartypeDecorHintPep557Exception):
        get_pep557_dataclass_kwargs(
            'I know the covert, from thence came I hither."')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **common configurations** (i.e., global :class:`beartype.BeartypeConf`
singletons providing frequently required default configurations leveraged
throughout the remainder of the public :mod:`beartype` API, typically as the
default values of optional keyword-only ``conf`` parameters).

This private submodule is *not* intended for direct importation by downstream
callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._conf.confmain import BeartypeConf
from beartype._conf.confenum import BeartypeStrategy
from beartype._util.cache.utilcachecall import callable_cached

# ....................{ GLOBALS                            }....................
BEARTYPE_CONF_DEFAULT = BeartypeConf()
'''
**Default beartype configuration** (i.e., :class:`BeartypeConf` class
instantiated with *no* parameters and thus default parameters), globalized to
trivially optimize external access to this configuration throughout this
codebase.

This global is intentionally *not* publicized to end users, who can simply
instantiate ``BeartypeConf()`` to efficiently obtain the same singleton.
'''

# ....................{ GETTERS                            }....................
@callable_cached
def get_beartype_conf_strategy_on() -> BeartypeConf:
    '''
    **Linear-time beartype configuration** (i.e., :class:`BeartypeConf` class
    instantiated with only a single parameter enabling :math:`O(n)` linear time
    complexity and otherwise with default parameters), globalized to trivially
    optimize external access to this configuration throughout this codebase.

    This configuration is intentionally exposed indirectly through this memoized
    getter rather than directly through a global singleton. While cumbersome,
    doing so marginally reduces the cost of importing this submodule for end
    users *not* importing a public :mod:`beartype` API calling this getter
    (e.g., the :mod:`beartype.door` subpackage).
    '''

    # Piercing through the frozen eternity of one-liners.
    return BeartypeConf(strategy=BeartypeStrategy.On)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`-, :pep:`585`-, and :pep:`646`-compliant **tuple type
hint utilities** (i.e., low-level callables generically applicable to
:pep:`484`- and :pep:`585`-compliant purely fixed- and variadic-length tuple
type hints *and* :pep:`646`-compliant mixed fixed-variadic tuple type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import Tuple
from beartype._data.typing.datatypingport import (
    Hint,
    TupleHints,
)
from beartype._data.hint.sign.datahintsigncls import HintSign
from beartype._data.hint.sign.datahintsigns import (
    HintSignPep484585TupleFixed,
    HintSignPep484585TupleVariadic,
    HintSignPep646TupleFixedVariadic,
)
from beartype._data.hint.sign.datahintsignset import (
    HINT_SIGNS_PEP646_TUPLE_HINT_CHILD_UNPACKED)
from beartype._util.hint.pep.proposal.pep484.pep484 import (
    HINT_PEP484_TUPLE_EMPTY)
from beartype._util.hint.pep.proposal.pep585 import (
    HINT_PEP585_TUPLE_EMPTY)

# ....................{ TESTERS                            }....................
def is_hint_pep484585646_tuple_variadic(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed object is either a :pep:`484`-, :pep:`585`-,
    or :pep:`646`-compliant **variable-length tuple hint** (i.e., parent hint
    subscripted by exactly two child hints, the former of which is any arbitrary
    hint and the latter of which is an **ellipsis** (i.e., the :data:`Ellipsis`
    singleton specified in code as an unquoted ``...``)).

    This tester returns :data:`True` if this hint is either:

    * A :pep:`484`-compliant variable-length tuple hint of the form
      ``typing.Tuple[{hint_child}, ...]`` for any ``{hint_child}``.
    * A :pep:`585`-compliant variable-length tuple hint of the form
      ``tuple[{hint_child}, ...]`` for any ``{hint_child}``.
    * A :pep:`646`-compliant unpacked variable-length child tuple hint of the
      form ``*tuple[{hint_child}, ...]`` for any ``{hint_child}``.

    This tester is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as this tester trivially reduces to an
    efficient one-liner.

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a variable-length tuple hint.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepget import get_hint_pep_args

    # Child hints subscripting this tuple hint.
    hint_childs = get_hint_pep_args(hint)

    # Return true only if...
    return (
        # This parent tuple hint is subscripted by two child hints *AND*...
        len(hint_childs) == 2 and
        # This second child hint is the PEP 484- and 585-compliant ellipsis
        # singleton (e.g., the unquoted character sequence "..." in
        # "tuple[str, ...]").
        hint_childs[1] is Ellipsis
    )


def is_hint_pep484585646_tuple_empty(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed object is either a :pep:`484`- or
    :pep:`585`-compliant **empty fixed-length tuple type hint** (i.e., hint
    constraining objects to be the empty tuple).

    Since type variables are not themselves types but rather placeholders
    dynamically replaced with types by type checkers according to various
    arcane heuristics, both type variables and types parametrized by type
    variables warrant special-purpose handling.

    This tester is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as this tester trivially reduces to an
    efficient one-liner and, in any case, is only called under rare edge cases.

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is an empty fixed-length tuple hint.
    '''

    # Return true only if this hint resembles either the PEP 484- or
    # 585-compliant fixed-length empty tuple type hint. Since there only exist
    # two such hints *AND* comparison against these hints is mostly fast, this
    # test is efficient in the general case.
    #
    # Note that this test may also be inefficiently performed by explicitly
    # obtaining this hint's sign and then subjecting this hint to specific
    # tests conditionally depending on which sign and thus PEP this hint
    # complies with: e.g.,
    #     # Return true only if this hint is either...
    #     return true (
    #         # A PEP 585-compliant "tuple"-based hint subscripted by no
    #         # child hints *OR*...
    #         (
    #             hint.__origin__ is tuple and
    #             hint_childs_len == 0
    #         ) or
    #         # A PEP 484-compliant "typing.Tuple"-based hint subscripted
    #         # by exactly one child hint *AND* this child hint is the
    #         # empty tuple,..
    #         (
    #             hint.__origin__ is Tuple and
    #             hint_childs_len == 1 and
    #             hint_childs[0] == ()
    #         )
    #     )
    return (
        hint == HINT_PEP585_TUPLE_EMPTY or
        hint == HINT_PEP484_TUPLE_EMPTY
    )

# ....................{ DISAMBIGUATORS                     }....................
#FIXME: Unit test us up, please.
def disambiguate_hint_pep484585646_tuple_sign(hint: Hint) -> HintSign:
    '''
    Disambiguate the passed **tuple type hint** (i.e., :pep:`484`- or
    :pep:`585`-compliant purely fixed- and variable-length tuple type hint *or*
    :pep:`646`-compliant mixed fixed-variadic tuple type hint) ambiguously
    identified by the :data:`.HintSignTuple` sign into whichever of the
    unambiguous :data:`.HintSignPep484585TupleFixed`,
    :data:`.HintSignPep484585TupleVariadic`, or
    :data:`.HintSignPep646TupleFixedVariadic` signs uniquely identify this kind
    of tuple type hint.

    This low-level getter assists the higher-level
    :func:`beartype._util.hint.pep.utilpepget.get_hint_pep_sign_or_none` getter
    to disambiguate the originally ambiguous :data:`.HintSignTuple` sign.

    This low-level getter is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as only function calling this getter is the
    aforementioned higher-level getter, which itself is memoized.

    Caveats
    -------
    **This getter exhibits worst-case** :math:`O(n)` **linear time complexity**
    for :math:`n` the number of child hints subscripting this tuple hint. Since
    this low-level getter is only called by higher-level memoized getters, this
    getter nonetheless effectively exhibits amortized worst-case :math:`O(1)`
    constant time complexity.

    Parameters
    ----------
    hint : Hint
        Tuple type hint to be disambiguated.

    Returns
    -------
    HintSign
        Sign uniquely and unambiguously identifying this hint. Specifically, if
        this hint is a:

        * :pep:`484`- or :pep:`585`-compliant **fixed-length tuple hint** (e.g.,
          of the form ``tuple[{hint_child_1}, ..., {hint_child_N}]``), this
          getter returns :data:`.HintSignPep484585TupleFixed`.
        * :pep:`484`- or :pep:`585`-compliant **variable-length tuple hint**
          (e.g., of the form ``tuple[{hint_child}, ...]``), this getter returns
          :data:`.HintSignPep484585TupleVariadic`.
        * :pep:`646`-compliant **fixed-variable tuple hint** (e.g., of the form
          ``tuple[{hint_child_1}, ..., {type_var_tuple}, ...,
          {hint_child_N}]``), this getter returns
          :data:`.HintSignPep646TupleFixedVariadic`.
    '''
    # print(f'Disambiguating tuple hint {repr(hint)}...')

    # ....................{ IMPORTS                        }....................
    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepget import get_hint_pep_args
    from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none

    # ....................{ LOCALS                         }....................
    # Child hints subscripting this parent tuple hint.
    hint_childs = get_hint_pep_args(hint)
    # print(f'hint_childs: {hint_childs}')

    # Number of child hints subscripting this parent tuple hint.
    hint_childs_len = len(hint_childs)

    # ....................{ PEP (484|585) ~ variadic       }....................
    # If this parent tuple hint is subscripted by *NO* child hints, this hint is
    # the unsubscripted "typing.Tuple" type hint factory semantically equivalent
    # to the PEP 484-compliant variable-length tuple hint "typing.Tuple[object,
    # ...]". In this case, return the sign uniquely identifying these hints.
    if not hint_childs_len:
        return HintSignPep484585TupleVariadic
    # Else, this parent tuple hint is subscripted by one or more child hints.

    # ....................{ PEP 646                        }....................
    # For each child hint subscripting this parent tuple hint...
    for hint_child in hint_childs:
        # Sign uniquely identifying this child hint if this child hint is
        # PEP-compliant *OR* "None" otherwise.
        hint_child_sign = get_hint_pep_sign_or_none(hint_child)
        # print(f'Detected hint child {hint_child} sign {hint_child_sign}...')

        # If this child hint is either:
        # * A PEP 646-compliant unpacked type variable tuple *OR*...
        # * A PEP 646-compliant unpacked child tuple hint...
        # ...then this parent tuple hint is PEP 646-compliant. In this case,
        # return the sign uniquely identifying these hints.
        if hint_child_sign in HINT_SIGNS_PEP646_TUPLE_HINT_CHILD_UNPACKED:
            return HintSignPep646TupleFixedVariadic
        # Else, this child hint is PEP 484- or 585-compliant.
    # Else, all child hints subscripting this parent tuple hint are *ONLY* PEP
    # 484- or 585-compliant, implying this parent tuple hint to itself be PEP
    # 484- or 585-compliant.

    # ....................{ PEP (484|585) ~ variadic       }....................
    # Return the sign uniquely identifying either...
    return (
        # Variable-length tuple hints if this hint is variadic;
        HintSignPep484585TupleVariadic
        if is_hint_pep484585646_tuple_variadic(hint) else
        # Fixed-length tuple hints otherwise.
        HintSignPep484585TupleFixed
    )

# ....................{ FACTORIES                          }....................
#FIXME: Unit test us up, please.
def make_hint_pep484585_tuple_fixed(hints_child: TupleHints) -> Hint:
    '''
    Dynamically create and return a new :pep:`484`- or :pep:`585`-compliant
    **fixed-length tuple hint** of the form ``tuple[{hint_child1}, ...,
    {hint_childN}]`` subscripted by all child hints in the passed tuple.

    Parameters
    ----------
    hints_child : TupleHints
        Tuple of all child hints with which to subscript the returned hint.

    Returns
    -------
    Hint
        Fixed-length tuple hint subscripted by these child hints.
    '''
    assert isinstance(hints_child, tuple), f'{repr(hints_child)} not tuple.'

    # Return a fixed-length tuple type hint subscripted by these child hints,
    # defined as either...
    return (
        #FIXME: Uncomment after dropping Python <= 3.10 support, which raises a
        #"SyntaxError" if we even try doing this. *SADNESS*
        # # If the active Python interpreter targets Python >= 3.11 and thus
        # # supports list unpacking in arbitrary expressions, prefer an efficient
        # # expression leveraging a list unpacking;
        # Tuple[*hints]
        # if IS_PYTHON_AT_LEAST_3_11 else
        # Else, the active Python interpreter targets Python <= 3.10.
        #
        # Dynamically subscript the builtin "tuple" type.
        Tuple.__class_getitem__(hints_child)  # type: ignore[attr-defined]
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`696`-compliant **type hint test data.**
'''

# ....................{ FIXTURES                           }....................
def hints_pep696_meta() -> 'List[HintPepMetadata]':
    '''
    List of :pep:`696`-compliant **type hint metadata** (i.e.,
    :class:`beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata`
    instances describing test-specific :pep:`696`-compliant sample type hints
    with metadata generically leveraged by various PEP-agnostic unit tests).
    '''

    # ..................{ IMPORTS                            }..................
    # Defer fixture-specific imports.
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_13

    # List of all PEP-specific type hint metadata to be returned.
    hints_pep_meta = []

    # If the active Python interpreter targets Python < 3.13, this interpreter
    # fails to support PEP 696. In this case, return the empty list.
    if not IS_PYTHON_AT_LEAST_3_13:
        return hints_pep_meta
    # Else, the active Python interpreter targets Python >= 3.13 and thus
    # supports PEP 696.

    # ..................{ IMPORTS ~ version                  }..................
    # Defer version-specific imports.
    from beartype.typing import (
        List,
        TypeVar,
    )
    from beartype.door import TypeVarTypeHint
    from beartype._data.hint.sign.datahintsigns import HintSignTypeVar
    from beartype_test.a00_unit.data.hint.util.data_hintmetacls import (
        HintPepMetadata,
        HintPithSatisfiedMetadata,
        HintPithUnsatisfiedMetadata,
    )

    # ..................{ LOCALS                             }..................
    T_list_str = TypeVar('T_list_str', default=List[str])
    '''
    '''

    # ..................{ LISTS                              }..................
    # Add PEP 696-specific test type hints to this list.
    hints_pep_meta.extend((
        # ................{ PEP 484 ~ typevar                  }................
        # PEP 484-compliant user-defined defaulted type variable.
        HintPepMetadata(
            hint=T_list_str,
            pep_sign=HintSignTypeVar,
            typehint_cls=TypeVarTypeHint,
            is_typing=False,
            piths_meta=(
                # List of string constants.
                HintPithSatisfiedMetadata(
                    ["And touch'd", 'with shade of', 'bronzed obelisks,',]),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    "Glar'd a blood-red through all its thousand courts,"),
            ),
        ),
    ))

    # ..................{ RETURN                             }..................
    # Return this list of all PEP-specific type hint metadata.
    return hints_pep_meta



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`692`-compliant **unpacked typed dictionary reducers** (i.e.,
low-level callables converting ``typing.Unpack[...]`` type hints subscripted by
:pep:`692`-compliant :class:`typing.TypedDict` subclasses to lower-level type
hints more readily consumable by :mod:`beartype`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
#FIXME: [PEP 692] Actually implement deep type-checking support for PEP
#692-compliant unpack type hints of the form "**kwargs:
#typing.Unpack[UserTypedDict]". Doing so will *ALMOST CERTAINLY* necessitate a
#new logic pathway for dynamically generating type-checking code efficiently
#type-checking the passed variadic keyword argument dictionary "**kwargs"
#against that user-defined "UserTypedDict". Feasible, but non-trivial. *sigh*

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep692Exception
from beartype.typing import Optional
from beartype._check.metadata.hint.hintsane import (
    HINT_SANE_IGNORABLE,
    HintSane,
)
from beartype._data.typing.datatypingport import Hint
from beartype._util.func.arg.utilfuncargiter import ArgKind

# ....................{ REDUCERS                           }....................
#FIXME: Split into disparate reducers:
#* A new reduce_hint_pep646_unpacked_typevartuple() reducer targeting
#  "HintSignPep646TypeVarTupleUnpacked".
#* A new reduce_hint_pep692() reducer targeting
#  "HintSignPep692TypedDictUnpacked".
#
#For the moment, both should continue reducing to "object". Oh! Wait. Right. We
#need to return "HINT_SANE_IGNORABLE" instead now, right? Make it so, please.
def reduce_hint_pep692(
    hint: Hint,
    arg_kind: Optional[ArgKind],
    exception_prefix: str,
    **kwargs
) -> HintSane:
    '''
    Reduce the passed :pep:`692`-compliant **unpacked typed dictionary** (i.e.,
    hint of the form "typing.Unpack[{typeddict}]" where "{typeddict}" is a
    :class:`typing.TypedDict` subclass) to a more readily digestible hint.

    This reducer effectively ignores this hint by reduction to the ignorable
    :class:`object` superclass (e.g., from ``**kwargs:
    typing.Unpack[UserTypedDict]`` to simply ``**kwargs``). Although non-ideal,
    generating code type-checking these hints is sufficiently non-trivial to
    warrant a (hopefully) temporary delay in doing so properly.

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as reducers cannot be memoized.

    Parameters
    ----------
    hint : Hint
        :pep:`646`- or :pep:`692`-compliant unpack type hint to be reduced.
    arg_kind : Optional[ArgKind]
        Either:

        * If this hint annotates a parameter of some callable, that parameter's
          **kind** (i.e., :class:`.ArgKind` enumeration member conveying the
          syntactic class of that parameter, constraining how the callable
          declaring that parameter requires that parameter to be passed).
        * Else, :data:`None`.
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    All remaining keyword-only parameters are silently ignored.

    Returns
    -------
    Hint
        Lower-level type hint currently supported by :mod:`beartype`.

    Raises
    ------
    BeartypeDecorHintPep692Exception
        If this hint annotates a variadic keyword parameter but is *not*
        subscripted by a single :pep:`589`-compliant :class:`typing.TypedDict`
        subclass.
    '''

    # If this hint does *NOT* directly annotate a variadic keyword parameter,
    # this hint is PEP 692-noncompliant. In this case, raise an exception.
    if arg_kind is not ArgKind.VARIADIC_KEYWORD:
        raise BeartypeDecorHintPep692Exception(
            f'{exception_prefix}PEP 692 unpacked typed dictionary {repr(hint)} '
            f'does not annotate variadic keyword parameter (i.e., callable '
            f'parameter of type {repr(arg_kind)} not variadic keyword).'
        )
        # Else, this child hint is a PEP 589-compliant "typing.TypeDict"
        # subclass.
    # Else, this hint directly annotates a variadic keyword parameter.

    # Silently reduce to a noop by returning this ignorable singleton global.
    # While non-ideal, worky is preferable to non-worky.
    return HINT_SANE_IGNORABLE



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype** :pep:`593` **utility unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.proposal.pep593` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ validators                 }....................
def test_die_unless_hint_pep593() -> None:
    '''
    Test the
    :beartype._util.hint.pep.proposal.pep593.die_unless_hint_pep593` validator.
    '''

    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep593Exception
    from beartype.typing import (
        Annotated,
        Optional,
    )
    from beartype._util.hint.pep.proposal.pep593 import die_unless_hint_pep593
    from pytest import raises

    # Assert this validator avoids raising exceptions for a type hint
    # subscripting this factory.
    die_unless_hint_pep593(Annotated[Optional[str], int])

    # Assert this validator raises the expected exception for an arbitrary
    # PEP-compliant type hint *NOT* subscripting this factory.
    with raises(BeartypeDecorHintPep593Exception):
        die_unless_hint_pep593(Optional[str])

# ....................{ TESTS ~ tester                     }....................
def test_is_hint_pep593_beartype() -> None:
    '''
    Test usage of the private
    :mod:`beartype._util.hint.pep.proposal.pep593.is_hint_pep593_beartype`
    tester.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPepException
    from beartype.vale import Is
    from beartype._util.hint.pep.proposal.pep593 import is_hint_pep593_beartype
    from pytest import raises
    from typing import Annotated

    # ....................{ PASS                           }....................
    # Assert this tester rejects beartype-agnostic metahints.
    assert is_hint_pep593_beartype(
        Annotated[str, 'And may there be no sadness of farewell']) is False

    # Assert this tester accepts a beartype-specific metahint annotated by a
    # beartype validator defined as a lambda function.
    assert is_hint_pep593_beartype(
        Annotated[str, Is[lambda text: bool(text)]]) is True

    # ....................{ FAIL                           }....................
    # Assert this tester raises the expected exception when passed a
    # non-metahint in either case.
    with raises(BeartypeDecorHintPepException):
        is_hint_pep593_beartype('When I embark;')

# ....................{ TESTS ~ getters                    }....................
def test_get_hint_pep593_metadata() -> None:
    '''
    Test the
    :beartype._util.hint.pep.proposal.pep593.get_hint_pep593_metadata`
    getter.
    '''

    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep593Exception
    from beartype.typing import (
        Annotated,
        Optional,
    )
    from beartype._util.hint.pep.proposal.pep593 import get_hint_pep593_metadata
    from pytest import raises

    # Assert this getter returns the expected tuple for an arbitrary PEP
    # 593-compliant type hint.
    assert get_hint_pep593_metadata(Annotated[
        Optional[str],
        'Thy', 'caverns', 'echoing', 'to', 'the', "Arve's", 'commotion,'
    ]) == (
        'Thy', 'caverns', 'echoing', 'to', 'the', "Arve's", 'commotion,')

    # Assert this getter raises the expected exception for an arbitrary
    # PEP-compliant type hint *NOT* subscripting this factory.
    with raises(BeartypeDecorHintPep593Exception):
        get_hint_pep593_metadata(Optional[str])


def test_get_hint_pep593_metahint() -> None:
    '''
    Test the
    :beartype._util.hint.pep.proposal.pep593.get_hint_pep593_metahint`
    getter.
    '''

    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep593Exception
    from beartype.typing import (
        Annotated,
        Optional,
    )
    from beartype._util.hint.pep.proposal.pep593 import get_hint_pep593_metahint
    from pytest import raises

    # Assert this getter returns the expected PEP-compliant type hint for an
    # arbitrary PEP 593-compliant type hint.
    metahint = Optional[int]
    assert get_hint_pep593_metahint(Annotated[
        metahint,
        'A', 'loud', 'lone', 'sound', 'no', 'other', 'sound', 'can', 'tame'
    ]) is metahint

    # Assert this getter raises the expected exception for an arbitrary
    # PEP-compliant type hint *NOT* subscripting this factory.
    with raises(BeartypeDecorHintPep593Exception):
        get_hint_pep593_metahint(metahint)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype import hookable** :pep:`526` **exception submodule** (i.e., data
module containing *only* :pep:`526`-compliant annotated variable assignments
raising :class:`beartype.roar.BeartypeDoorHintViolation` violations, mimicking
real-world usage of the :func:`beartype.claw.beartype_package` import hook from
external callers).
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDoorHintViolation
from beartype.typing import (
    List,
    Union,
)
from beartype._util.text.utiltextrepr import represent_object
from pytest import raises

# Import another submodule of this subpackage implicitly installing different
# import hooks configured by another beartype configuration isolated to that
# submodule. Doing so exercises that beartype import hooks correctly support
# import hook composition (i.e., combinations of arbitrary import hooks).
from beartype_test.a00_unit.data.claw.intraprocess.hookable_package.beartype_this_package import (
    this_submodule)

# from beartype.claw._clawstate import claw_state
# print(f'this_submodule conf: {repr(claw_state.module_name_to_beartype_conf)}')

# ....................{ PEP 526                            }....................
# Validate that the import hook presumably installed by the caller implicitly
# appends all PEP 526-compliant annotated assignment statements in this
# submodule with calls to beartype's statement-level
# beartype.door.die_if_unbearable() exception-raiser.

# Assert that a PEP 526-compliant annotated assignment statement assigning an
# object satisfying the type hint annotating that statement raises *NO*
# exception.
and_winter_robing: int = len('And winter robing with pure snow and crowns')

# Assert that a PEP 526-compliant annotated statement lacking an assignment
# raises *NO* exception.
such_magic_as_compels_the_charmed_night: str

# Arbitrary object violating type hints below.
OF_STARRY_ICE = len('Of starry ice the grey grass and bare boughs;')

# Assert that a PEP 526-compliant annotated assignment statement assigning an
# object violating the type hint annotating that statement raises the expected
# exception.
with raises(BeartypeDoorHintViolation) as exception_info:
    of_starry_ice: Union[float, List[str]] = OF_STARRY_ICE

# Exception message raised by that assignment.
exception_message = str(exception_info.value)

# Truncated representation of the invalid object assigned by that assignment.
pith_repr = represent_object(OF_STARRY_ICE)

# Assert that this message contains a truncated representation of this pith.
assert pith_repr in exception_message

# Assert that this raiser successfully replaced the temporary
# placeholder previously prefixing this message.
assert 'global ' in exception_message.lower()
assert __name__ in exception_message
assert ' violates type hint ' in exception_message

# ....................{ CLASSES                            }....................
class ThePausesOfHerMusic(object):
    and_her_breath: int = 'The pauses of her music, and her breath'
    '''
    Class variable whose initial value violates the :pep:`526`-compliant type
    hint annotating this variable.

    This variable effectively validates that the import hook presumably
    installed by the caller ignores all :pep:`526`-compliant annotated class
    variable assignment statements by *not* implicitly appending these
    statements with calls to the statement-level
    :func:`beartype.door.die_if_unbearable` exception-raiser.
    '''

    pass

# ....................{ CALLABLES                          }....................
def beneath_the_cold_glare() -> None:
    '''
    Arbitrary callable whose body contains one or more :pep:`526`-compliant
    annotated variable assignments, exercising edge cases in
    :mod:`beartype.claw` import hooks handling these assignments.
    '''

    # Arbitrary object violating type hints below.
    OF_THE_DESOLATE_NIGHT = len('Beneath the cold glare of the desolate night,')

    # Assert that a PEP 526-compliant annotated assignment statement assigning
    # an object violating the type hint annotating that statement raises the
    # expected exception.
    with raises(BeartypeDoorHintViolation) as exception_info:
        through_tangled_swamps: Union[float, List[str]] = OF_THE_DESOLATE_NIGHT

    # Exception message raised by that assignment.
    local_exception_message = str(exception_info.value)

    # Truncated representation of the invalid object assigned by that assignment.
    local_pith_repr = represent_object(OF_THE_DESOLATE_NIGHT)

    # Assert that this message contains a truncated representation of this pith.
    assert local_pith_repr in local_exception_message

    # Assert that this raiser successfully replaced the temporary
    # placeholder previously prefixing this message.
    assert 'callable ' in local_exception_message.lower()
    assert __name__ in local_exception_message
    assert 'beneath_the_cold_glare() ' in local_exception_message
    assert ' violates type hint ' in local_exception_message



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **PEP-noncompliant NumPy type hint** (i.e., type hint defined by
the third-party :mod:`numpy` package) utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: The top-level of this module should avoid importing from third-party
# optional libraries, both because those libraries cannot be guaranteed to be
# either installed or importable here *AND* because those imports are likely to
# be computationally expensive, particularly for imports transitively importing
# C extensions (e.g., anything from NumPy or SciPy).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **PEP-agnostic type hint sanitizers** (i.e., high-level callables
converting type hints from one format into another, either permanently or
temporarily and either losslessly or in a lossy manner).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import Optional
from beartype._cave._cavemap import NoneTypeOr
from beartype._check.convert._convcoerce import (
    coerce_func_hint_root,
    coerce_hint_root,
)
from beartype._check.convert._reduce.redmain import reduce_hint
from beartype._check.metadata.hint.hintsane import HintSane
from beartype._check.metadata.metadecor import BeartypeDecorMeta
from beartype._conf.confmain import BeartypeConf
from beartype._conf.confcommon import BEARTYPE_CONF_DEFAULT
from beartype._data.error.dataerrmagic import EXCEPTION_PLACEHOLDER
from beartype._data.func.datafuncarg import ARG_NAME_RETURN
from beartype._data.hint.sign.datahintsigncls import HintSign
from beartype._data.kind.datakindiota import SENTINEL
from beartype._data.typing.datatypingport import Hint
from beartype._data.typing.datatyping import (
    HintSignOrNoneOrSentinel,
    TypeStack,
)
from beartype._util.func.arg.utilfuncargiter import ArgKind
from beartype._util.hint.pep.proposal.pep484585.pep484585func import (
    reduce_hint_pep484585_func_return)

# ....................{ SANIFIERS ~ root                   }....................
#FIXME: Unit test us up, please.
def sanify_hint_root_func(
    # Mandatory parameters.
    decor_meta: BeartypeDecorMeta,
    hint: Hint,
    pith_name: str,

    # Optional parameters.
    arg_kind: Optional[ArgKind] = None,
    exception_prefix: str = EXCEPTION_PLACEHOLDER,
) -> HintSane:
    '''
    Type hint sanified (i.e., sanitized) from the passed **possibly insane root
    type hint** (i.e., possibly PEP-noncompliant hint annotating the parameter
    or return with the passed name of the passed callable) if this hint is both
    reducible and unignorable, this hint unmodified if this hint is both
    irreducible and unignorable, or :data:`.HINT_SANE_IGNORABLE` otherwise (i.e., if
    this hint is ignorable).

    Specifically, this function:

    * If this hint is a **PEP-noncompliant tuple union** (i.e., tuple of one or
      more standard classes and forward references to standard classes):

      * Coerces this tuple union into the equivalent :pep:`484`-compliant
        union.
      * Replaces this tuple union in the ``__annotations__`` dunder tuple of
        this callable with this :pep:`484`-compliant union.
      * Returns this :pep:`484`-compliant union.

    * Else if this hint is already PEP-compliant, preserves and returns this
      hint unmodified as is.
    * Else (i.e., if this hint is neither PEP-compliant nor -noncompliant and
      thus invalid as a type hint), raise an exception.

    Caveats
    -------
    This sanifier *cannot* be meaningfully memoized, since the passed type hint
    is *not* guaranteed to be cached somewhere. Only functions passed cached
    type hints can be meaningfully memoized. Even if this function *could* be
    meaningfully memoized, there would be no benefit; this function is only
    called once per parameter or return of the currently decorated callable.

    This sanifier is intended to be called *after* all possibly
    :pep:`563`-compliant **deferred type hints** (i.e., type hints persisted as
    evaluatable strings rather than actual type hints) annotating this callable
    if any have been evaluated into actual type hints.

    Parameters
    ----------
    decor_meta : BeartypeDecorMeta
        Decorated callable directly annotated by this hint.
    hint : Hint
        Possibly PEP-noncompliant root type hint to be sanified.
    pith_name : str
        Either:

        * If this hint annotates a parameter, the name of that parameter.
        * If this hint annotates the return, ``"return"``.
    arg_kind : Optional[ArgKind]
        Either:

        * If this hint annotates a parameter, that parameter's **kind** (i.e.,
          :class:`.ArgKind` enumeration member conveying the syntactic class of
          that parameter, constraining how the callable declaring that parameter
          requires that parameter to be passed).
        * If this hint annotates the return, :data:`None`.

        Defaults to :data:`None`.
    exception_prefix : str, optional
        Human-readable substring prefixing raised exception messages. Defaults
        to :data:`.EXCEPTION_PLACEHOLDER`.

    Returns
    -------
    HintSane
        Either:

        * If this hint is ignorable, :data:`.HINT_SANE_IGNORABLE`.
        * Else if this unignorable hint is reducible to another hint, metadata
          encapsulating this reduction.
        * Else, this unignorable hint is irreducible. In this case, metadata
          encapsulating this hint unmodified.

    Raises
    ------
    BeartypeDecorHintNonpepException
        If this object is neither:

        * A PEP-noncompliant type hint.
        * A supported PEP-compliant type hint.
    '''
    assert isinstance(arg_kind, NoneTypeOr[ArgKind]), (
        f'{repr(arg_kind)} neither argument kind nor "None".')

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Synchronize with the sanify_hint_root_statement() sanitizer.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    # PEP-compliant hint coerced from this possibly PEP-noncompliant hint if
    # this hint is coercible *OR* this hint as is otherwise. Since the passed
    # hint is *NOT* necessarily PEP-compliant, perform this coercion *BEFORE*
    # validating this hint to be PEP-compliant.
    hint_coerced = coerce_func_hint_root(
        decor_meta=decor_meta,
        hint=hint,
        pith_name=pith_name,
        exception_prefix=exception_prefix,
    )

    # If this possibly PEP-noncompliant hint was actually coerced into a
    # PEP-compliant hint...
    if hint_coerced is not hint:
        # Note this coercion.
        hint = hint_coerced

        # Safely set the hint annotating the parameter or return with the passed
        # name of the decorated callable to the passed hint in a portable manner
        # consistent with both PEP 649 and Python >= 3.14.
        decor_meta.set_func_pith_hint(pith_name=pith_name, hint=hint)
    # Else, this possibly PEP-noncompliant hint was *NOT* coerced into a
    # PEP-compliant hint, implying this hint to already be PEP-compliant.

    # If this hint annotates the return, then (in order):
    # * If this hint is contextually invalid for this callable (e.g., generator
    #   whose return is not annotated as "Generator[...]"), raise an exception.
    # * If this hint is either PEP 484- or 585-compliant *AND* requires
    #   reduction (e.g., from "Coroutine[None, None, str]" to just "str"),
    #   reduce this hint accordingly.
    #
    # Perform this reduction *BEFORE* performing subsequent tests (e.g., to
    # accept "Coroutine[None, None, typing.NoReturn]" as expected). Note that
    # this logic *ONLY* pertains to callables (rather than statements) and is
    # thus *NOT* performed by the sanify_hint_root_statement() sanitizer.
    if pith_name == ARG_NAME_RETURN:
        hint = reduce_hint_pep484585_func_return(
            func=decor_meta.func_wrappee,
            func_annotations=decor_meta.func_annotations,
            exception_prefix=exception_prefix,
        )
    # Else, this hint annotates a parameter.

    # Sane child hint reduced from this possibly insane child hint if reducing
    # this hint did not generate supplementary metadata *OR* that metadata
    # otherwise (i.e., if reducing this hint generated supplementary metadata).
    # Reductions simplify subsequent logic elsewhere by transparently converting
    # non-trivial hints (e.g., numpy.typing.NDArray[...]) into semantically
    # equivalent trivial hints (e.g., beartype validators).
    #
    # Whereas the above coercion permanently persists for the duration of the
    # active Python process (i.e., by replacing the original type hint in the
    # annotations dunder dictionary of this callable), this reduction only
    # temporarily persists for the duration of the current call stack. Why?
    # Because hints explicitly coerced above are assumed to be either:
    # * PEP-noncompliant and thus harmful (in the general sense).
    # * PEP-compliant but semantically deficient and thus equally harmful (in
    #   the general sense).
    #
    # In either case, coerced type hints are generally harmful in *ALL* possible
    # contexts for *ALL* possible consumers (including other competing runtime
    # type-checkers). Reduced type hints, however, are *NOT* harmful in any
    # sense whatsoever; they're simply non-trivial for @beartype to support in
    # their current form and thus temporarily reduced in-memory into a more
    # convenient form for beartype-specific type-checking elsewhere.
    hint_sane = reduce_hint(
        hint=hint,
        conf=decor_meta.conf,
        decor_meta=decor_meta,
        arg_kind=arg_kind,
        cls_stack=decor_meta.cls_stack,
        pith_name=pith_name,
        exception_prefix=exception_prefix,
    )

    # Return this hint if this hint is unignorable *OR* "typing.Any" otherwise.
    return hint_sane


#FIXME: Unit test us up, please.
def sanify_hint_root_statement(
    hint: Hint,
    conf: BeartypeConf,
    exception_prefix: str,
) -> HintSane:
    '''
    PEP-compliant type hint sanified (i.e., sanitized) from the passed **root
    type hint** (i.e., possibly PEP-noncompliant type hint that has *no* parent
    type hint) if this hint is both reducible and unignorable, this hint
    unmodified if this hint is both irreducible and unignorable, or
    :data:`.HINT_SANE_IGNORABLE` otherwise (i.e., if this hint is ignorable).

    This sanifier is principally intended to be called by a **statement-level
    type-checker factory** (i.e., a function creating and returning a runtime
    type-checker type-checking this hint, outside the context of any standard
    type hinting annotation like a user-defined class variable, callable
    parameter or return, or assignment statement). Such factories include:

    * The private :func:`beartype._check.checkmake.make_func_tester` factory,
      internally called by:

      * The public :func:`beartype.door.die_if_unbearable` function.
      * The public :func:`beartype.door.is_bearable` function.
      * The public :meth:`beartype.door.TypeHint.die_if_unbearable` method.
      * The public :meth:`beartype.door.TypeHint.is_bearable` method.

    Parameters
    ----------
    hint : Hint
        Possibly PEP-noncompliant root type hint to be sanified.
    conf : BeartypeConf
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object).
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    Returns
    -------
    HintSane
        Either:

        * If this hint is ignorable, :data:`.HINT_SANE_IGNORABLE`.
        * Else if this unignorable hint is reducible to another hint, metadata
          encapsulating this reduction.
        * Else, this unignorable hint is irreducible. In this case, metadata
          encapsulating this hint unmodified.

    Raises
    ------
    BeartypeDecorHintNonpepException
        If this object is neither:

        * A PEP-noncompliant type hint.
        * A supported PEP-compliant type hint.

    See Also
    --------
    :func:`.sanify_hint_root_func`
        Further details.
    '''

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Synchronize with the sanify_hint_root_func() sanitizer, please.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    # PEP-compliant type hint coerced from this possibly PEP-noncompliant type
    # hint if this hint is coercible *OR* this hint as is otherwise. Since the
    # passed hint is *NOT* necessarily PEP-compliant, perform this coercion
    # *BEFORE* validating this hint to be PEP-compliant.
    hint = coerce_hint_root(hint=hint, exception_prefix=exception_prefix)

    # Metadata encapsulating the sanification of this hint.
    hint_sane = reduce_hint(
        hint=hint, conf=conf, exception_prefix=exception_prefix)

    # Return this metadat.
    return hint_sane

# ....................{ SANIFIERS ~ any                    }....................
#FIXME: This function accepting a "pith_name" parameter is *SUPER-WEIRD.* If
#this function did *NOT* accept a "pith_name" parameter, then the implementation
#could internally reduce to calling the safer reduce_hint_child() function.
#Instead, the current implementation has *NO* choice but to call the less safe
#reduce_hint() function. Clearly, the docstring below suggests this has
#something suspicious to do with error-handling. Investigate, please. *sigh*
#FIXME: Unit test us up, please.
def sanify_hint_child(
    # Mandatory parameters.
    hint: Hint,
    hint_parent_sane: Optional[HintSane],

    # Optional parameters.
    cls_stack: TypeStack = None,
    conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
    hint_sign_seed: HintSignOrNoneOrSentinel = SENTINEL,
    pith_name: Optional[str] = None,
    exception_prefix: str = '',
) -> HintSane:
    '''
    Metadata encapsulating the sanification (i.e., sanitization) of the passed
    **possibly insane child type hint** (i.e., possibly PEP-noncompliant hint
    transitively subscripting the root hint annotating a parameter or return of
    the currently decorated callable) if this hint is both reducible and
    unignorable, this hint unmodified if this hint is both irreducible and
    unignorable, or :obj:`.HINT_SANE_IGNORABLE` otherwise (i.e., if this hint is
    ignorable).

    Parameters
    ----------
    hint : Hint
        Child type hint to be sanified.
    hint_parent_sane : Optional[HintSane]
        Either:

        * If this hint is actually a **root type hint,** :data:`None`.
        * Else, **Sanified parent type hint metadata** (i.e., immutable and thus
          hashable object encapsulating *all* metadata previously returned by
          :mod:`beartype._check.convert.convmain` sanifiers after sanitizing
          the possibly PEP-noncompliant parent hint of this child hint into a
          fully PEP-compliant parent hint).
    cls_stack : TypeStack, default: None
        **Type stack** (i.e., either a tuple of the one or more
        :func:`beartype.beartype`-decorated classes lexically containing the
        class variable or method annotated by this hint *or* :data:`None`).
        Defaults to :data:`None`.
    conf : BeartypeConf, default: BEARTYPE_CONF_DEFAULT
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object). Defaults
        to :obj:`.BEARTYPE_CONF_DEFAULT`, the default beartype configuration.
    hint_sign_seed :  HintSignOrNoneOrSentinel, default: SENTINEL
        **Type hint seed sign** (i.e., sign identifying this hint with respect
        to the first reduction performed by this sanification) if this hint is
        ambiguously identifiable by two or more signs *or* the sentinel
        otherwise (i.e., if this hint is uniquely identifiable by one sign).

        This sign is used to seed (i.e., initialize) the first reduction
        internally performed by this sanification, which otherwise defaults to
        the sign returned by the :func:`.get_hint_pep_sign_or_none` getter. This
        parameter should only be passed to handle edge cases in which a hint is
        ambiguously identifiable by two or more signs, including:

        * **Typed dictionary generics** (i.e., user-defined types subclassing
          both the :pep:`484`-compliant :class:`typing.Generic` superclass and
          :pep:`589`-compliant :class:`typing.TypedDict` superclass), which are
          identifiable as both generics *and* typed dictionaries.

        Defaults to the sentinel.
    pith_name : Optional[str], default: None
        Either:

        * If this hint directly annotates a callable parameter (as the root type
          hint of that parameter), the name of this parameter.
        * If this hint directly annotates a callable return (as the root type
          hint of that return), the magic string ``"return"``.
        * Else, :data:`None`.

        Note that:

        * This parameter should only be passed during exception raising (i.e.,
          from within the :func:`beartype._check.error` subpackage).
        * This parameter should *never* be passed during code generation (i.e.,
          by the :func:`beartype._check.code.codemain.make_check_expr` code
          factory).

        Defaults to :data:`None`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    HintSane
        Either:

        * If this hint is ignorable, :data:`.HINT_SANE_IGNORABLE`.
        * Else if this unignorable hint is reducible to another hint, metadata
          encapsulating this reduction.
        * Else, this unignorable hint is irreducible. In this case, metadata
          encapsulating this hint unmodified.
    '''
    # print(f'Sanifying child hint {repr(hint)} with type variable lookup table {repr(typearg_to_hint)}...')

    # This sanifier covers the proper subset of logic performed by the
    # sanify_hint_root_statement() sanifier applicable to child type hints.
    #
    # Note that this subset is currently the lower-level reduce_hint() reducer.
    # Technically, this implies that this sanifier could simply be a trivial
    # alias of that reducer. Pragmatically, doing so would only negligibly
    # improve decoration-time speed (which is *SORTA* good) while demonstrably
    # harming code maintainability (which is *DEFINITELY* bad). Ergo, we
    # intentionally maintain this sanifier as a distinct and separate function.
    # You'll thank me ten years from now, older self. *high five*

    # Metadata encapsulating the sanification of this child hint.
    #
    # Note that this function intentionally performs *NO* hint coercion (e.g.,
    # by calling the coerce_hint_any() coercer). Although feasible, doing so
    # would be pointless. Why? Because hint coercion coerces unmemoized hints
    # into memoized hints as a means of deduplicating hints across
    # "__annotations__" dunder dictionaries. However, there exists *NO*
    # "__annotations__" dunder dictionary to deduplicate across here. There
    # exists *NO* space to be compacted here and thus *NO* demonstrable
    # reason to perform hint coercion.
    hint_sane = reduce_hint(
        hint=hint,
        hint_parent_sane=hint_parent_sane,
        hint_sign_seed=hint_sign_seed,
        conf=conf,
        cls_stack=cls_stack,
        pith_name=pith_name,
        exception_prefix=exception_prefix,
    )
    # print(f'[sanify] Detecting hint {repr(hint)} reduction {repr(hint_sane)} ignorability...')

    # Return this metadata.
    return hint_sane


def sanify_hint_any(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    hint_parent_sane: Optional[HintSane] = None,
    **kwargs
) -> HintSane:
    '''
    Metadata encapsulating the sanification (i.e., sanitization) of the passed
    **possibly insane type hint** (i.e., possibly PEP-noncompliant hint
    transitively subscripting the root hint annotating a parameter or return of
    the currently decorated callable) if this hint is both reducible and
    unignorable, this hint unmodified if this hint is both irreducible and
    unignorable, or :obj:`.HINT_SANE_IGNORABLE` otherwise (i.e., if this hint is
    ignorable).

    Caveats
    -------
    **The more fine-grained** :func:`.sanify_hint_child` **sanifier should
    typically be called instead.** This more coarse-grained sanifier drops the
    mandatory ``hint_parent_sane`` parameter required by the former, which is
    *not* necessarily a good thing. That parameter should typically be passed.
    Failing to pass that parameter drops essential metadata required to properly
    sanitize many insane type hints.

    Parameters
    ----------
    hint : Hint
        Child type hint to be sanified.
    hint_parent_sane : Optional[HintSane], default: None
        Either:

        * If this hint is actually a **root type hint,** :data:`None`.
        * Else, **Sanified parent type hint metadata** (i.e., immutable and thus
          hashable object encapsulating *all* metadata previously returned by
          :mod:`beartype._check.convert.convmain` sanifiers after sanitizing
          the possibly PEP-noncompliant parent hint of this child hint into a
          fully PEP-compliant parent hint).

        Defaults to :data:`None`.

    All remaining keyword parameters are as accepted by the comparable
    :func:`.sanify_hint_child` sanifier.

    Returns
    -------
    HintSane
        Either:

        * If this hint is ignorable, :data:`.HINT_SANE_IGNORABLE`.
        * Else if this unignorable hint is reducible to another hint, metadata
          encapsulating this reduction.
        * Else, this unignorable hint is irreducible. In this case, metadata
          encapsulating this hint unmodified.
    '''

    # Defer to our betters.
    return sanify_hint_child(
        hint=hint, hint_parent_sane=hint_parent_sane, **kwargs)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **wrapper function code snippets** (i.e., triple-quoted pure-Python
string constants formatted and concatenated together to dynamically generate the
implementations of wrapper functions type-checking
:func:`beartype.beartype`-decorated callables).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.code.datacodename import (
    ARG_NAME_ARGS_NAME_KEYWORDABLE,
    ARG_NAME_FUNC,
    ARG_NAME_GET_VIOLATION,
    VAR_NAME_ARGS_LEN,
    VAR_NAME_PITH_ROOT,
)
from beartype._data.code.datacodeindent import CODE_INDENT_1
from beartype._data.typing.datatyping import CallableStrFormat
from beartype._util.func.arg.utilfuncargiter import ArgKind

# ....................{ CODE                               }....................
CODE_SIGNATURE = f'''{{code_signature_prefix}}def {{func_name}}(
    *args,
{{code_signature_scope_args}}{CODE_INDENT_1}**kwargs
):'''
'''
Code snippet declaring the signature of a type-checking callable.

Note that the :func:`beartype._check.signature.make_signature` factory function
internally interpolates these format variables into this string as follows:

* ``code_signature_prefix`` is replaced by:

  * For synchronous callables, the empty string.
  * For asynchronous coroutines (but *not* asynchronous generators, curiously),
    the space-suffixed keyword ``"async "``.

* ``code_signature_scope_args`` is replaced by a comma-delimited string listing
  all :mod:`beartype`-specific hidden parameters internally required to
  type-check the currently decorated callable.
'''


CODE_INIT_ARGS_LEN = f'''
    # Localize the number of passed positional arguments for efficiency.
    {VAR_NAME_ARGS_LEN} = len(args)'''
'''
Code snippet localizing the number of passed positional arguments for callables
accepting one or more such arguments.
'''

# ....................{ CODE ~ arg                         }....................
ARG_KIND_TO_CODE_LOCALIZE = {
    # Snippet localizing any positional-only parameter (e.g.,
    # "{posonlyarg}, /") by lookup in the wrapper's "*args" dictionary.
    ArgKind.POSITIONAL_ONLY: f'''
    # If this positional-only parameter was passed...
    if {VAR_NAME_ARGS_LEN} > {{arg_index}}:
        # Localize this positional-only parameter.
        {VAR_NAME_PITH_ROOT} = args[{{arg_index}}]''',

    # Snippet localizing any positional or keyword parameter as follows:
    #
    # * If this parameter's 0-based index (in the parameter list of the
    #   decorated callable's signature) does *NOT* exceed the number of
    #   positional parameters passed to the wrapper function, localize this
    #   positional parameter from the wrapper's variadic "*args" tuple.
    # * Else if this parameter's name is in the dictionary of keyword
    #   parameters passed to the wrapper function, localize this keyword
    #   parameter from the wrapper's variadic "*kwargs" tuple.
    # * Else, this parameter is unpassed. In this case, localize this parameter
    #   as a placeholder value guaranteed to *NEVER* be passed to any wrapper
    #   function: the private "__beartypistry" singleton passed to this wrapper
    #   function as a hidden default parameter and thus accessible here. While
    #   we could pass a "__beartype_sentinel" parameter to all wrapper
    #   functions defaulting to "object()" and then use that here instead,
    #   doing so would slightly reduce efficiency for no tangible gain. *shrug*
    ArgKind.POSITIONAL_OR_KEYWORD: f'''
    # Localize this positional or keyword parameter if passed *OR* to the
    # sentinel "__beartype_raise_exception" guaranteed to never be passed.
    {VAR_NAME_PITH_ROOT} = (
        args[{{arg_index}}] if {VAR_NAME_ARGS_LEN} > {{arg_index}} else
        kwargs.get({{arg_name!r}}, {ARG_NAME_GET_VIOLATION})
    )

    # If this parameter was passed...
    if {VAR_NAME_PITH_ROOT} is not {ARG_NAME_GET_VIOLATION}:''',

    # Snippet localizing any keyword-only parameter (e.g., "*, {kwarg}") by
    # lookup in the wrapper's variadic "**kwargs" dictionary. (See above.)
    ArgKind.KEYWORD_ONLY: f'''
    # Localize this keyword-only parameter if passed *OR* to the sentinel value
    # "__beartype_raise_exception" guaranteed to never be passed.
    {VAR_NAME_PITH_ROOT} = kwargs.get({{arg_name!r}}, {ARG_NAME_GET_VIOLATION})

    # If this parameter was passed...
    if {VAR_NAME_PITH_ROOT} is not {ARG_NAME_GET_VIOLATION}:''',

    #FIXME: [SPEED] Optimize this from a "for" into "while" loop, please.
    #"while" loops internally raise *NO* "StopException" whereas "for" loops do.
    #Snippet iteratively localizing all variadic positional parameters. *sigh*
    ArgKind.VARIADIC_POSITIONAL: f'''
    # For all excess positional parameters in the passed "*args" parameter...
    for {VAR_NAME_PITH_ROOT} in args[{{arg_index!r}}:]:''',

    #FIXME: [SPEED] Optimize this from a "for" into "while" loop. See above!
    # Snippet iteratively localizing all variadic keyword parameters.
    ArgKind.VARIADIC_KEYWORD: f'''
    # For all excess keyword parameters in the passed "**kwargs" parameter,
    # decided by subtracting the subset of all keywordable parameters
    # explicitly accepted by this callable from the set of all parameters passed
    # by keyword to this callable...
    for {VAR_NAME_PITH_ROOT} in (
        (kwargs[kwarg_name] for kwarg_name in kwargs.keys() - {ARG_NAME_ARGS_NAME_KEYWORDABLE})):''',
}
'''
Dictionary mapping from the type of each callable parameter supported by the
:func:`beartype.beartype` decorator to a code snippet localizing that callable's
next parameter to be type-checked.
'''

# ....................{ CODE ~ return ~ check              }....................
CODE_RETURN_CHECK_PREFIX = f'''
    # Call this function with all passed parameters and localize the value
    # returned from this call.
    {VAR_NAME_PITH_ROOT} = {{func_call_prefix}}{ARG_NAME_FUNC}(*args, **kwargs)

    # Noop required to artificially increase indentation level. Note that
    # CPython implicitly optimizes this conditional away. Isn't that nice?
    if True:'''
'''
Code snippet calling the decorated callable and localizing the value returned by
that call.

Note that:

* The :func:`beartype._decor._nontype._wrap.wrapmaingenerate_code` factory
  function internally interpolates these format variables into this string as
  follows:

  * ``func_call_prefix`` is replaced by:

    * For synchronous callables, the empty string.
    * For asynchronous coroutine factories (but *not* asynchronous generator
      factories, curiously), the space-suffixed keyword ``"await "``.

* This snippet intentionally terminates on a noop increasing the indentation
  level, enabling subsequent type-checking code to effectively ignore
  indentation level and thus uniformly operate on both:

  * Parameters localized via values of the
    :data:`.PARAM_KIND_TO_PEP_CODE_LOCALIZE` dictionary.
  * Return values localized via this snippet.

See Also
--------
https://stackoverflow.com/a/18124151/2809027
    Bytecode disassembly demonstrating that CPython optimizes away the spurious
   ``if True:`` conditional hardcoded into this snippet.
'''


CODE_RETURN_CHECK_SUFFIX = f'''
    {{func_return_prefix}}{VAR_NAME_PITH_ROOT}'''
'''
Code snippet returning from the wrapper function the successfully type-checked
value returned from the decorated callable.

Note that the :func:`beartype._decor._nontype._wrap.wrapmaingenerate_code`
factory function internally interpolates these format variables into this string
as follows:

* ``func_return_prefix`` is replaced by:

  * For synchronous generator factories (but *not* asynchronous generator
    factories, curiously), the space-suffixed keyword ``"yield from "``.
  * For all other callables, the space-suffixed keyword ``"return "``.
'''

# ....................{ CODE ~ return ~ uncheck            }....................
CODE_RETURN_UNCHECKED = f'''
    # Call this function with all passed parameters and return the value
    # returned from this call as is (without being type-checked).
    {{func_return_prefix}}{{func_call_prefix}}{ARG_NAME_FUNC}(*args, **kwargs)'''
'''
Code snippet calling the decorated (either synchronous or non-synchronous)
non-generator callable *without* type-checking the value returned by that call
(if any).

See Also
--------
:data:`.CODE_RETURN_CHECK_PREFIX`
:data:`.CODE_RETURN_CHECK_SUFFIX`
    Further details on format variables.
'''

# ..................{ FORMATTERS                             }..................
# str.format() methods, globalized to avoid inefficient dot lookups elsewhere.
# This is an absurd micro-optimization. *fight me, github developer community*
CODE_RETURN_CHECK_PREFIX_format: CallableStrFormat = (
    CODE_RETURN_CHECK_PREFIX.format)
CODE_RETURN_CHECK_SUFFIX_format: CallableStrFormat = (
    CODE_RETURN_CHECK_SUFFIX.format)
CODE_RETURN_UNCHECKED_format: CallableStrFormat = (
    CODE_RETURN_UNCHECKED.format)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **declarative object validation classes** (i.e.,
:mod:`beartype`-specific classes enabling callers to define PEP-compliant
validators from arbitrary caller-defined objects tested via explicitly
supported object introspectors efficiently generating stack-free code).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
# All "FIXME:" comments for this submodule reside in this package's "__init__"
# submodule to improve maintainability and readability here.

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeValeSubscriptionException
from beartype.typing import Any, Tuple
from beartype.vale._is._valeisabc import _BeartypeValidatorFactoryABC
from beartype.vale._util._valeutilsnip import (
    VALE_CODE_CHECK_ISATTR_TEST_format,
    VALE_CODE_CHECK_ISATTR_VALUE_EXPR_format,
    VALE_CODE_INDENT_1,
)
from beartype.vale._core._valecore import BeartypeValidator
from beartype._data.typing.datatyping import LexicalScope
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.kind.maplike.utilmapset import update_mapping
from beartype._util.func.utilfuncscope import add_func_scope_attr
from beartype._util.text.utiltextrepr import represent_object
from beartype._data.kind.datakindiota import SENTINEL

# ....................{ SUBCLASSES ~ attr                  }....................
class _IsAttrFactory(_BeartypeValidatorFactoryABC):
    '''
    **Beartype object attribute validator factory** (i.e., object creating and
    returning a new beartype validator when subscripted (indexed) by both the
    name of any object attribute *and* any **attribute validator** (i.e., other
    beartype validator created by subscripting any :mod:`beartype.vale` class),
    validating that :mod:`beartype`-decorated callable parameters and returns
    annotated by :attr:`typing.Annotated` type hints subscripted by the former
    validator define an attribute with that name satisfying that attribute
    validator).

    This class efficiently validates that callable parameters and returns
    define arbitrary object attributes satisfying arbitrary validators
    subscripting this factory. Any :mod:`beartype`-decorated callable parameter
    or return annotated by a :attr:`typing.Annotated` type hint subscripted by
    this factory subscripted by any object attribute name and validator (e.g.,
    ``typing.Annotated[{cls}, beartype.vale.IsAttr[{attr_name},
    {attr_validator}]]`` for any class ``{cls}``, object attribute name
    ``{attr_name}`, and object attribute validator ``{attr_validator}``)
    validates that parameter or return value to be an instance of that class
    defining an attribute with that name satisfying that attribute validator.

    **This factory incurs no time performance penalties at call time.** Whereas
    the general-purpose :class:`beartype.vale.Is` factory necessarily calls
    the caller-defined callable subscripting that factory at call time and thus
    incurs a minor time performance penalty, this factory efficiently reduces
    to one-line tests in :mod:`beartype`-generated wrapper functions *without*
    calling any callables and thus incurs *no* time performance penalties.

    Examples
    --------
    .. code-block:: pycon

       # Import the requisite machinery.
       >>> from beartype import beartype
       >>> from beartype.vale import IsAttr, IsEqual
       >>> from typing import Annotated
       >>> import numpy as np

       # Type hint matching only two-dimensional NumPy arrays of 64-bit floats,
       # generating code resembling:
       #    (isinstance(array, np.ndarray) and
       #     array.ndim == 2 and
       #     array.dtype == np.dtype(np.float64))
       >>> Numpy2dFloat64Array = Annotated[
       ...     np.ndarray,
       ...     IsAttr['ndim', IsEqual[2]],
       ...     IsAttr['dtype', IsEqual[np.dtype(np.float64)]],
       ... ]

       # Type hint matching only one-dimensional NumPy arrays of 64-bit floats,
       # generating code resembling:
       #    (isinstance(array, np.ndarray) and
       #     array.ndim == 2 and
       #     array.dtype.type == np.float64)
       >>> Numpy1dFloat64Array = Annotated[
       ...     np.ndarray,
       ...     IsAttr['ndim', IsEqual[2]],
       ...     # Nested attribute validators test equality against a "."-delimited
       ...     # attribute lookup (e.g., "dtype.type"), as expected.
       ...     IsAttr['dtype', IsAttr['type', IsEqual[np.float64]]],
       ... ]

       # NumPy arrays of well-known real number series.
       >>> FAREY_2D_FLOAT64_ARRAY = np.array(
       ...     [[0/1, 1/8,], [1/7, 1/6,], [1/5, 1/4], [2/7, 1/3], [3/8, 2/5]])
       >>> FAREY_1D_FLOAT64_ARRAY = np.array(
       ...     [3/7, 1/2, 4/7, 3/5, 5/8, 2/3, 5/7, 3/4, 4/5, 5/6, 6/7, 7/8])

       # Annotate callables by those type hints.
       >>> @beartype
       ... def sqrt_sum_2d(
       ...     array: Numpy2dFloat64Array) -> Numpy1dFloat64Array:
       ...     """
       ...     One-dimensional NumPy array of 64-bit floats produced by first
       ...     summing the passed two-dimensional NumPy array of 64-bit floats
       ...     along its second dimension and then square-rooting those sums.
       ...     """
       ...     return np.sqrt(array.sum(axis=1))

       # Call those callables with parameters satisfying those hints.
       >>> sqrt_sum_2d(FAREY_2D_FLOAT64_ARRAY)
       [0.35355339 0.55634864 0.67082039 0.78679579 0.88034084]

       # Call those callables with parameters violating those hints.
       >>> sqrt_sum_2d(FAREY_1D_FLOAT64_ARRAY)
       beartype.roar.BeartypeCallHintParamViolation: @beartyped
       sqrt_sum_2d() parameter array="array([0.42857143, 0.5, 0.57142857, 0.6,
       0.625, ...])" violates type hint typing.Annotated[numpy.ndarray,
       IsAttr['ndim', IsEqual[2]], IsAttr['dtype', IsEqual[dtype('float64')]]],
       as value "array([0.42857143, 0.5, 0.57142857, 0.6, 0.625, ...])"
       violates validator IsAttr['ndim', IsEqual[2]].

    See Also
    --------
    :class:`beartype.vale.Is`
        Further commentary.
    '''

    # ..................{ DUNDERS                            }..................
    @callable_cached
    def __getitem__(  # type: ignore[override]
        self, args: Tuple[str, BeartypeValidator]) -> BeartypeValidator:
        '''
        Create and return a new beartype validator validating object attributes
        with the passed name satisfying the passed validator, suitable for
        subscripting :pep:`593`-compliant :attr:`typing.Annotated` type hints.

        This method is memoized for efficiency.

        Parameters
        ----------
        args : Tuple[str, BeartypeValidator]
            2-tuple ``(attr_name, attr_validator)``, where:

            * ``attr_name`` is the arbitrary attribute name to validate that
              parameters and returns define satisfying the passed validator.
            * ``attr_validator`` is the attribute validator to validate that
              attributes with the passed name of parameters and returns
              satisfy.

        Returns
        -------
        BeartypeValidator
            Beartype validator encapsulating this validation.

        Raises
        ------
        BeartypeValeSubscriptionException
            If this factory was subscripted by either:

            * *No* arguments.
            * One argument.
            * Three or more arguments.

        See Also
        ----------
        :class:`._IsAttrFactory`
            Usage instructions.
        '''

        # If this class was subscripted by one non-tuple argument, raise an
        # exception.
        if not isinstance(args, tuple):
            raise BeartypeValeSubscriptionException(
                f'{self._getitem_exception_prefix}non-tuple argument '
                f'{represent_object(args)}.'
            )
        # Else, this class was subscripted by either no *OR* two or more
        # arguments (contained in this tuple).
        #
        # If this class was *NOT* subscripted by two arguments...
        elif len(args) != 2:
            # If this class was subscripted by one or more arguments, then by
            # deduction this class was subscripted by three or more arguments.
            # In this case, raise a human-readable exception.
            if args:
                raise BeartypeValeSubscriptionException(
                    f'{self._getitem_exception_prefix}three or more arguments '
                    f'{represent_object(args)}.'
                )
            # Else, this class was subscripted by *NO* arguments. In this case,
            # raise a human-readable exception.
            else:
                raise BeartypeValeSubscriptionException(
                    f'{self._getitem_exception_prefix}empty tuple.')
        # Else, this class was subscripted by exactly two arguments.

        # Localize these arguments to human-readable local variables.
        attr_name, attr_validator = args

        # Representer (i.e., callable accepting *NO* arguments returning a
        # machine-readable representation of this validator), defined *AFTER*
        # localizing these validator arguments.
        get_repr = lambda: (
            f'beartype.vale.{self._basename}['
            f'{repr(attr_name)}, {repr(attr_validator)}'
            f']'
        )

        # If this name is *NOT* a string, raise an exception.
        if not isinstance(attr_name, str):
            raise BeartypeValeSubscriptionException(
                f'{get_repr()} first argument '
                f'{represent_object(attr_name)} not string.'
            )
        # Else, this name is a string.
        #
        # If this name is the empty string, raise an exception.
        elif not attr_name:
            raise BeartypeValeSubscriptionException(
                f'{get_repr()} first argument is empty string.')
        # Else, this name is a non-empty string.
        #
        # Note that this name has *NOT* yet been validated to be valid Python
        # identifier. While we could do so here by calling our existing
        # is_identifier() tester, doing so would inefficiently repeat
        # the split on "." characters performed below. Instead, we iteratively
        # validate each split substring to be a valid Python identifier below.

        # Callable inefficiently validating object attributes with this name
        # against this validator.
        # is_valid: BeartypeValidatorTester = None  # type: ignore[assignment]

        # Code snippet efficiently validating object attributes with this name
        # against this validator.
        is_valid_code = ''

        # Dictionary mapping from the name to value of each local attribute
        # referenced in the "is_valid_code" snippet defined below.
        is_valid_code_locals: LexicalScope = {}

        # If this attribute name is unqualified (i.e., contains no "."
        # delimiters), prefer an efficient optimization avoiding iteration.
        if '.' not in attr_name:
            # If this name is *NOT* a valid Python identifier, raise an
            # exception.
            if not attr_name.isidentifier():
                raise BeartypeValeSubscriptionException(
                    f'{get_repr()} first argument {repr(attr_name)} not '
                    f'syntactically valid Python identifier.'
                )
            # Else, this name is a valid Python identifier.

            def is_valid(pith: Any) -> bool:
                f'''
                :data:`True` only if the passed object defines an attribute
                named "{attr_name}" whose value satisfies the validator
                {repr(attr_validator)}.
                '''

                # Attribute of this object with this name if this object
                # defines such an attribute *OR* a sentinel placeholder
                # otherwise (i.e., if this object defines *NO* such attribute).
                pith_attr = getattr(pith, attr_name, SENTINEL)

                # Return true only if...
                return (
                    # This object defines an attribute with this name *AND*...
                    pith_attr is not SENTINEL and
                    # This attribute satisfies this validator.
                    attr_validator.is_valid(pith_attr)
                )

            # Names of new parameters added to the signature of wrapper
            # functions enabling this validator to be tested in those functions
            # *WITHOUT* additional stack frames whose values are:
            # * The sentinel placeholder.
            #
            # Add these parameters *BEFORE* generating locals.
            local_name_sentinel = add_func_scope_attr(
                attr=SENTINEL, func_scope=is_valid_code_locals)

            # Generate locals safely merging the locals required by both the
            # code generated below *AND* the externally provided code
            # validating this attribute.
            update_mapping(
                mapping_trg=is_valid_code_locals,
                mapping_src=attr_validator._is_valid_code_locals,
            )

            #FIXME: Unfortunately, "local_name_attr_value" still isn't a
            #sufficiently unique name below, because "IsAttr['name',
            #IsAttr['name', IsEqual[True]]]" is a trivial counter-example where
            #the current approach breaks down. For true uniquification here,
            #we're going to need to instead:
            #* Define a global private counter:
            #  _local_name_obj_attr_value_counter = Counter(0)
            #* Replace the assignment below with:
            #  local_name_obj_attr_value = (
            #      f'{{obj}}_isattr_'
            #      f'{next(_local_name_obj_attr_value_counter)}'
            #  )
            #
            #Of course, this assumes "Counter" objects are thread-safe. If
            #they're not, we'll need to further obfuscate all this behind a
            #[R]Lock of some sort. *sigh*
            #FIXME: Oh, right. We mixed up "collections.Counter" with
            #"itertools.count". The former is orthogonal to our interests here;
            #the latter is of interest but *NOT* thread-safe. The solution is
            #for us to implement a new "FastWriteCounter" class resembling that
            #published in this extremely clever (and thus awesome) article:
            #    https://julien.danjou.info/atomic-lock-free-counters-in-python

            # Name of a local variable in this code whose:
            # * Name is sufficiently obfuscated as to be hopefully unique to
            #   the code generated by this validator.
            # * Value is the value of this attribute of the arbitrary object
            #   being validated by this code.
            local_name_attr_value = f'{{obj}}_isattr_{attr_name}'

            #FIXME: *OVERKILL.* The "VALE_CODE_CHECK_ISATTR_VALUE_EXPR" and
            #"VALE_CODE_CHECK_ISATTR_TEST" globals are *ONLY* ever accessed in
            #this specific method. Refactor as follows, please:
            #* Merge the entirety of the lower-level
            #  "VALE_CODE_CHECK_ISATTR_VALUE_EXPR" substring into the parent
            #  "VALE_CODE_CHECK_ISATTR_TEST" substring.
            #* Remove the "VALE_CODE_CHECK_ISATTR_VALUE_EXPR" and
            #  "_VALE_CODE_CHECK_ISATTR_VALUE_EXPR_RAW" globals entirely.

            # Python expression expanding to the value of this attribute,
            # efficiently optimized under Python >= 3.8 with an assignment
            # expression to avoid inefficient access of this value.
            attr_value_expr = VALE_CODE_CHECK_ISATTR_VALUE_EXPR_format(
                #FIXME: Inefficient and unnecessary. Since "attr_name" is
                #guaranteed to be a valid Python identifier, the
                #"VALE_CODE_CHECK_ISATTR_VALUE_EXPR" template should just
                #forcefully embed "{attr_name_expr}" inside single quotes.
                attr_name_expr=repr(attr_name),
                local_name_attr_value=local_name_attr_value,
                local_name_sentinel=local_name_sentinel,
            )

            # Python expression validating the value of this attribute,
            # formatted so as to be safely embeddable in the larger code
            # expression defined below.
            attr_value_is_valid_expr = attr_validator._is_valid_code.format(
                # Replace the placeholder substring "{obj}" in this code with
                # the expression expanding to this attribute's value, defined as
                # the name of the local variable previously assigned the value
                # of this attribute by the "VALE_CODE_CHECK_ISATTR_VALUE_EXPR"
                # code snippet subsequently embedded in the
                # "VALE_CODE_CHECK_ISATTR_VALUE_TEST" code snippet.
                obj=local_name_attr_value,
                # Replace the placeholder substring "{indent}" in this code with
                # an indentation increased by one level.
                indent=VALE_CODE_INDENT_1,
            )

            # Code snippet efficiently validating against this object.
            is_valid_code = VALE_CODE_CHECK_ISATTR_TEST_format(
                attr_value_expr=attr_value_expr,
                attr_value_is_valid_expr=attr_value_is_valid_expr,
                local_name_sentinel=local_name_sentinel,
            )
        # Else, this attribute name is qualified (i.e., contains one or more
        # "." delimiters), fallback to a general solution performing iteration.
        else:
            #FIXME: Implement us up when we find the time, please. We currently
            #raise an exception simply because we ran out of time for this. :{
            raise BeartypeValeSubscriptionException(
                f'{get_repr()} first argument '
                f'{repr(attr_name)} not unqualified Python identifier '
                f'(i.e., contains one or more "." characters).'
            )

        # Create and return this subscription.
        return BeartypeValidator(
            is_valid=is_valid,
            is_valid_code=is_valid_code,
            is_valid_code_locals=is_valid_code_locals,
            get_repr=get_repr,
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator** :pep:`695`-compliant **type alias** unit tests.

This submodule unit tests :pep:`695` support implemented in the
:func:`beartype.beartype` decorator.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_if_python_version_less_than

# ....................{ TESTS                              }....................
@skip_if_python_version_less_than('3.12.0')
def test_decor_pep695_type_recursive_indirect() -> None:
    '''
    Test :pep:`695` support implemented in the :func:`beartype.beartype`
    decorator if the active Python interpreter targets Python >= 3.12 *or* skip
    otherwise.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype_test.a00_unit.data.pep.pep695.data_pep695util import (
        unit_test_decor_hint_pep695_type_recursive_indirect)

    # Perform this test.
    unit_test_decor_hint_pep695_type_recursive_indirect()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **callable globals** (i.e., global constants describing various
well-known functions and methods).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ SETS                               }....................
OBJECT_SLOT_WRAPPERS = frozenset(
    # This slow wrapper.
    object_attr_value
    # For the name of each attribute defined by the root "object" superclass...
    for object_attr_name, object_attr_value in object.__dict__.items()
    # If the value of this attribute is callable and thus a low-level C-based
    # slot wrapper whose default implementation is mostly useless.
    if callable(object_attr_value)
)
'''
Frozen set of all **object slot wrappers** (i.e., low-level C-based callables
bound to the root :class:`object` superclass providing mostly useless default
implementations of popular dunder methods).

The default implementations of object slot wrappers have no intrinsic value in
any meaningful context and only serve to obfuscate actually intrinsically
valuable methods declared by concrete subclasses. Detecting and ignoring object
slot wrappers is thus a common desire.
'''


METHOD_NAMES_DUNDER_BINARY = frozenset((
    '__add__',
    '__and__',
    '__cmp__',
    '__divmod__',
    '__div__',
    '__eq__',
    '__floordiv__',
    '__ge__',
    '__gt__',
    '__iadd__',
    '__iand__',
    '__idiv__',
    '__ifloordiv__',
    '__ilshift__',
    '__imatmul__',
    '__imod__',
    '__imul__',
    '__ior__',
    '__ipow__',
    '__irshift__',
    '__isub__',
    '__itruediv__',
    '__ixor__',
    '__le__',
    '__lshift__',
    '__lt__',
    '__matmul__',
    '__mod__',
    '__mul__',
    '__ne__',
    '__or__',
    '__pow__',
    '__radd__',
    '__rand__',
    '__rdiv__',
    '__rfloordiv__',
    '__rlshift__',
    '__rmatmul__',
    '__rmod__',
    '__rmul__',
    '__ror__',
    '__rpow__',
    '__rrshift__',
    '__rshift__',
    '__rsub__',
    '__rtruediv__',
    '__rxor__',
    '__sub__',
    '__truediv__',
    '__xor__',
))
'''
Frozen set of the unqualified names of all **binary dunder methods** (i.e.,
methods whose names are both prefixed and suffixed by ``__``, which the active
Python interpreter implicitly calls to perform binary operations on instances
whose first operands are instances of the classes declaring those methods).
'''



#!/usr/bin/env python3
#--------------------( LICENSE                             )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Decidedly Object-Oriented Runtime-checking (DOOR) callable type hint classes**
(i.e., :class:`beartype.door.TypeHint` subclasses implementing support for
:pep:`484`- and :pep:`585`-compliant ``Tuple[...]`` type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.door._cls.doorsuper import TypeHint
from beartype.door._cls.pep.pep484585.doorpep484585subscripted import (
    SubscriptedTypeHint)
from beartype._util.hint.pep.proposal.pep484585646 import (
    is_hint_pep484585646_tuple_empty,
    is_hint_pep484585646_tuple_variadic,
)

# ....................{ SUBCLASSES                         }....................
class TupleFixedTypeHint(TypeHint):
    '''
    **Fixed-length tuple type hint wrapper** (i.e., high-level object
    encapsulating a low-level :pep:`484`- or :pep:`585`-compliant type hint of
    the form ``tuple[{child_hint_1}, ..., {child_hint_N}]``).
    '''

    # ..................{ INITIALIZERS                       }..................
    def _make_args(self) -> tuple:

        # Tuple of the zero or more low-level child type hints subscripting
        # (indexing) the low-level parent type hint wrapped by this wrapper.
        args = super()._make_args()

        # If this is the empty fixed-length tuple type hint (e.g., "tuple[()]"),
        # reduce this awkward nested-empty-tuple-in-a-1-tuple to an elegant
        # empty tuple as expected by sane users. Even if users wanted us to
        # return an awkward nested-empty-tuple-in-a-1-tuple, we couldn't. Why?
        # Because an empty tuple is *NOT* otherwise a valid type hint and
        # *CANNOT* thus be wrapped by an instance of the "TypeHint" superclass.
        if is_hint_pep484585646_tuple_empty(self._hint):
            args = ()
        # Else, this is a non-empty fixed-length tuple type hint (e.g.,
        # "tuple[int, str]"). In this case, preserve these child hints as is.

        # Return these child hints.
        return args

    # ..................{ PRIVATE ~ properties               }..................
    @property
    def _is_args_ignorable(self) -> bool:

        # Prevent the child type hints subscripting *ANY* fixed-length tuple
        # type hints from being ignored as a whole, even if one or all of those
        # child type hints are technically ignorable. Why? Because ignoring all
        # child type hints as a whole would then prevent the length of this
        # tuple type hint from being type-checked, which would rather defeat the
        # purpose of the whole thing really.
        return False

    # ..................{ PRIVATE ~ testers                  }..................
    def _is_subhint_branch(self, branch: TypeHint) -> bool:

        # print(f'self: {repr(self)}; branch: {repr(branch)}')

        # If the other hint branch is unsubscripted (e.g., "typing.Callable"),
        # assume that hint to be subscripted as "typing.Callable[..., Any]" by
        # reducing to a test for compatible origin types.
        if branch._is_args_ignorable:
            return issubclass(self._origin, branch._origin)
        # Else, that hint is subscripted.
        #
        # If that hint is a variable-length tuple, then this fixed-length tuple
        # is commensurable (i.e., comparable) with that hint despite having a
        # differing class. In this case...
        elif isinstance(branch, TupleVariableTypeHint):
            # Child hint subscripting that variable-length tuple.
            branch_child = branch._args_wrapped_tuple[0]

            # Return true only if *ALL* child hints subscripting this
            # fixed-length tuple are subhints of the single child hint
            # subscripting that variable-length tuple: e.g.,
            #     >>> TypeHint(tuple[str, str]) <= TypeHint(tuple[str, ...])
            #     True
            #     >>> TypeHint(tuple[str, int]) <= TypeHint(tuple[str, ...])
            #     False
            return all(
                self_child <= branch_child
                for self_child in self._args_wrapped_tuple
            )
        # Else, the other hint is *NOT* a variable-length tuple.
        #
        # If the other hint is also *NOT* a fixed-length tuple, these two hints
        # are incommensurable. In this case, return false.
        elif not isinstance(branch, TupleFixedTypeHint):
            return False
        # Else, the other hint is also a fixed-length tuple.
        #
        # If these two hints are subscripted by a differing number of child
        # hints, return false.
        elif len(self) != len(branch):
            return False
        # Else, these two hints are subscripted by the same number of child
        # hints.

        # Return true only if each child hint subscripting this hint is a
        # subhint of the corresponding child hint subscripting that hint.
        return all(
            self_child <= branch_child
            for self_child, branch_child in zip(
                self._args_wrapped_tuple, branch._args_wrapped_tuple)
        )


class TupleVariableTypeHint(SubscriptedTypeHint):
    '''
    **Variable-length tuple type hint wrapper** (i.e., high-level object
    encapsulating a low-level :pep:`484`- or :pep:`585`-compliant type hint of
    the form ``tuple[{child_hint}, ...]``).
    '''

    # ..................{ INITIALIZERS                       }..................
    def _make_args(self) -> tuple:

        # Tuple of the two low-level child type hints subscripting (indexing)
        # the low-level parent type hint wrapped by this wrapper.
        args = super()._make_args()

        # Validate this parent type hint to be subscripted sanely.
        #
        # Note that the previously called get_hint_pep_sign() getter already
        # validated this to be the case.
        assert is_hint_pep484585646_tuple_variadic(self._hint), (
            f'PEP 585 variable-length tuple type hint {repr(self._hint)} '
            f'not of the form "tuple[{{child_hint}}, ...]".'
        )

        # Reduce this tuple to the 1-tuple consisting *ONLY* of the first child
        # type hint subscripting this parent type hint. Equivalently, ignore the
        # second child type subscripting this parent type hint; by the above
        # validation, the latter is guaranteed to be the ellipses singleton,
        # which is *NOT* otherwise a valid type hint and *CANNOT* thus be
        # wrapped by an instance of the "TypeHint" superclass.
        return (args[0],)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype Inferential Type-hint Engine (BITE) API.**

This subpackage provides a declarative API for intelligently inferring (i.e.,
reverse-engineering) PEP-compliant type hints from arbitrary caller-defined
objects.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid polluting the public module namespace, external attributes
# should be locally imported at module scope *ONLY* under alternate private
# names (e.g., "from argparse import ArgumentParser as _ArgumentParser" rather
# than merely "from argparse import ArgumentParser").
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype.bite._infermain import (
    infer_hint as infer_hint,
)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
High-level **beartype integration tests.**

This submodule functionally tests the this project's behaviour with respect to
imports of both internal subpackages and submodules (unique to this project)
*and* external third-party packages and modules.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_beartype_import_isolation() -> None:
    '''
    Integration test ensuring that importing the top-level lightweight
    :mod:`beartype` package does *not* accidentally import from one or more
    heavyweight (sub)packages or third-party packages.

    This integration test assists the :mod:`beartype` package minimize the fixed
    cost of its first importation by downstream consumers.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.py.utilpyinterpreter import (
        get_interpreter_command_words)
    from beartype_test._util.command.pytcmdrun import (
        run_command_forward_stderr_return_stdout)
    from re import (
        compile as re_compile,
        search as re_search,
    )

    # ....................{ LOCALS                         }....................
    #FIXME: *FRAGILE.* Manually hardcoding module names here invites
    #desynchronization, particularly with optional third-party dependencies.
    #Instead, the names of optional third-party packages should be dynamically
    #constructed from the contents of the "beartype.meta" submodule.

    # Tuple of uncompiled regular expressions, each matching the
    # fully-qualified name of *ANY* heavyweight (sub)module or package whose
    # importation violates our self-imposed constraint of fast importation of
    # our core @beartype.beartype decorator.
    #
    # Note that:
    # * The third-party "typing_extensions" module has been profiled across
    #   all supported CPython versions to import either faster or only slightly
    #   slower than the standard "typing" module. In either case, both modules
    #   implement sufficiently rapidly as to be ignorable with respect to
    #   importation costs here. See also @posita's stunning profiling work at:
    #       https://github.com/beartype/beartype/pull/103#discussion_r815027198
    HEAVY_MODULE_NAME_RAW_REGEXES = (
        r'beartype\.bite',
        r'beartype\.cave',
        r'beartype\.claw',
        r'beartype\.door',
        r'beartype\.peps',
        r'beartype\.vale',
        r'numpy',
    )

    # Uncompiled regular expressions synthesized from this tuple.
    HEAVY_MODULE_NAME_RAW_REGEX = '|'.join(HEAVY_MODULE_NAME_RAW_REGEXES)

    # Compiled regular expression synthesized from this tuple.
    HEAVY_MODULE_NAME_REGEX = re_compile(
        fr'^({HEAVY_MODULE_NAME_RAW_REGEX})\b')
    # print(f'_HEAVY_MODULE_NAME_REGEX: {_HEAVY_MODULE_NAME_REGEX}')

    # Python code snippet printing all imports (i.e., the contents of the
    # standard "sys.modules" list) *AFTER* importing the @beartype decorator.
    CODE_IMPORT_BEARTYPE_AND_PRINT_IMPORTED_MODULES = '''
# Import the core @beartype decorator and requisite "sys" machinery.
from beartype import beartype
from sys import modules

# Print a newline-delimited list of the fully-qualified names of all modules
# transitively imported by the prior "import" statements.
print('\\n'.join(module_name for module_name in modules.keys()))
'''

    # Tuple of all arguments to be passed to the active Python interpreter rerun
    # as an external command.
    PYTHON_ARGS = get_interpreter_command_words() + (
        '-c', CODE_IMPORT_BEARTYPE_AND_PRINT_IMPORTED_MODULES)

    # ....................{ PASS                           }....................
    # Run this code isolated to a Python subprocess, raising an exception on
    # subprocess failure while both forwarding all standard error output by this
    # subprocess to the standard error file handle of the active Python process
    # *AND* capturing and returning all subprocess stdout.
    module_names_str = run_command_forward_stderr_return_stdout(
        command_words=PYTHON_ARGS)

    #FIXME: Actually parse "module_names" for bad entries here, please.
    # print(module_names)

    # List of the fully-qualified names of all modules transitively imported by
    # importing the @beartype decorator in an isolated Python interpreter.
    module_names = module_names_str.splitlines()

    # For each such name, assert this name is *NOT* that of a heavyweight
    # (sub)module or package enumerated above.
    #
    # Note that this iteration could, of course, also be more efficiently
    # implementable as a single composite regex match on the newline-delimited
    # "module_names_str" string. However, doing so would entirely defeat the
    # purpose of this iteration: to offer unambiguous and human-readable error
    # messages in the common event of an importation violation.
    for module_name in module_names:
        assert re_search(HEAVY_MODULE_NAME_REGEX, module_name) is None, (
            f'@beartype.beartype improperly imports heavyweight module '
            f'"{module_name}" from global scope.'
        )


def test_beartype_optimized(monkeypatch: 'pytest.MonkeyPatch') -> None:
    '''
    Integration test ensuring that the :func:`beartype.beartype` decorator
    behaves as expected under **Python optimization** (e.g., by the
    CPython-specific ``PYTHONOPTIMIZE=1`` environment variable).

    Parameters
    ----------
    monkeypatch : MonkeyPatch
        :mod:`pytest` fixture allowing various state associated with the active
        Python process to be temporarily changed for the duration of this test.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.py.utilpyinterpreter import (
        get_interpreter_command_words)
    from beartype_test._util.command.pytcmdrun import (
        run_command_forward_output)

    # ....................{ LOCALS                         }....................
    # Python code snippet importing, configuring, and applying the @beartype
    # decorator to an arbitrary user-defined callable.
    CODE_OPTIMIZE_BEARTYPE = '''
# ....................{ IMPORTS                            }....................
from beartype import beartype, BeartypeConf
from beartype.roar import BeartypeCallHintParamViolation
from beartype.typing import Union
from beartype._util.py.utilpyinterpreter import is_python_optimized
from pytest import raises

# ....................{ MAIN                               }....................
towertype = beartype(conf=BeartypeConf(is_pep484_tower=True))
"""
Custom :func:`beartype.beartype` decorator configured with non-default options.
"""

@towertype
def towering_func(int_or_float: float) -> float:
    """
    Arbitrary callable decorated by the custom :func:`beartype.beartype`
    decorator defined above, validating that :func:`beartype.beartype` supports
    optimized Python interpreters in configuration mode.
    """

    # Make a gibberish mathematical operation transparently supporting both
    # integers and floating-point numbers so, Ensign Bear!
    return int_or_float ** int_or_float ^ round(int_or_float)


@beartype
def untowering_func(int_or_float: Union[int, float]) -> Union[int, float]:
    """
    Arbitrary callable decorated by the standard :func:`beartype.beartype`
    decorator imported above, validating that :func:`beartype.beartype` supports
    optimized Python interpreters in decoration mode.
    """

    # Make a gibberish mathematical operation transparently supporting both
    # integers and floating-point numbers so, Ensign Bear!
    return int_or_float ** int_or_float ^ round(int_or_float)

# ....................{ ASSERTS                            }....................
# Assert these callables when passed valid input return the expected result.
assert towering_func(3) == 24
assert untowering_func(3) == 24

# Type of exception expected to be raised by passing this callable invalid
# input, defined as either...
exception_type_expected = (
    # If the active Python interpreter is optimized, the expected low-level
    # builtin CPython runtime exception;
    TypeError
    if is_python_optimized() else
    # Else, the active Python interpreter is unoptimized. In this case, the
    # expected @beartype-specific type-checking violation.
    BeartypeCallHintParamViolation
)

# Assert these callables when passed invalid input raise the expected exception.
with raises(exception_type_expected):
    towering_func("Until he reach'd the great main cupola;")
with raises(exception_type_expected):
    untowering_func('There standing fierce beneath, he stampt his foot,')
'''

    # Tuple of all arguments to be passed to the active Python interpreter rerun
    # as an external command.
    PYTHON_ARGS = get_interpreter_command_words() + (
        '-c', CODE_OPTIMIZE_BEARTYPE,)

    # ....................{ PASS                           }....................
    # For all relevant values of the "${PYTHONOPTIMIZE}" environment variable,
    # where:
    # * A value of "0" disables Python optimization.
    # * A value of "1" enables the lowest level of Python optimization.
    for python_optimize_value in ('0', '1',):
        # Temporarily set the Python optimization state by setting the
        # "${PYTHONOPTIMIZE}" environment variable to this string value.
        monkeypatch.setenv('PYTHONOPTIMIZE', python_optimize_value)

        # Run this code isolated to a (possibly optimized) Python subprocess,
        # raising an exception on subprocess failure while both forwarding all
        # standard output and error output by this subprocess to the standard
        # output and error file handles of the active Python process.
        run_command_forward_output(command_words=PYTHON_ARGS)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **string-joining utility unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.text.utiltextjoin` submodule.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                             }....................
def test_join_delimited() -> None:
    '''
    Test the :func:`beartype._util.text.utiltextjoin.join_delimited` function.
    '''

    # Defer test-specific imports.
    from beartype._util.text.utiltextjoin import join_delimited

    # Assert that joining a sequence of no strings returns the empty string.
    assert join_delimited(
        strs=(),
        delimiter_if_two='In Xanadu did Kubla Khan',
        delimiter_if_three_or_more_nonlast='A stately pleasure-dome decree:',
        delimiter_if_three_or_more_last='Where Alph, the sacred river, ran',
    ) == ''

    # Assert that joining a sequence of one string returns that string.
    assert join_delimited(
        strs=('Through caverns measureless to man',),
        delimiter_if_two='Down to a sunless sea.',
        delimiter_if_three_or_more_nonlast=(
            'So twice five miles of fertile ground'),
        delimiter_if_three_or_more_last=(
            'With walls and towers were girdled round;'),
    ) == 'Through caverns measureless to man'

    # Assert that joining a sequence of two strings returns these strings
    # conditionally delimited by the appropriate delimiter.
    assert join_delimited(
        strs=(
            'And there were gardens bright with sinuous rills,',
            'Where blossomed many an incense-bearing tree;',
        ),
        delimiter_if_two='And here were forests ancient as the hills,',
        delimiter_if_three_or_more_nonlast=(
            'Enfolding sunny spots of greenery.'),
        delimiter_if_three_or_more_last=(
            'But oh! that deep romantic chasm which slanted'),
    ) == (
        'And there were gardens bright with sinuous rills,'
        'And here were forests ancient as the hills,'
        'Where blossomed many an incense-bearing tree;'
    )

    # Assert that joining a sequence of three strings returns these strings
    # conditionally delimited by the appropriate delimiters.
    assert join_delimited(
        strs=(
            'Down the green hill athwart a cedarn cover!',
            'A savage place! as holy and enchanted',
            'As e’er beneath a waning moon was haunted',
        ),
        delimiter_if_two='By woman wailing for her demon-lover!',
        delimiter_if_three_or_more_nonlast=(
            'And from this chasm, with ceaseless turmoil seething,'),
        delimiter_if_three_or_more_last=(
            'As if this earth in fast thick pants were breathing,'),
    ) == (
        'Down the green hill athwart a cedarn cover!'
        'And from this chasm, with ceaseless turmoil seething,'
        'A savage place! as holy and enchanted'
        'As if this earth in fast thick pants were breathing,'
        'As e’er beneath a waning moon was haunted'
    )


def test_join_delimited_disjunction() -> None:
    '''
    Test the
    :func:`beartype._util.text.utiltextjoin.join_delimited_disjunction`
    function.
    '''

    # Defer test-specific imports.
    from beartype._util.text.utiltextjoin import join_delimited_disjunction

    # Assert that joining a sequence of no strings returns the empty string.
    assert join_delimited_disjunction((
        'A mighty fountain momently was forced:',
        'Amid whose swift half-intermitted burst',
        'Huge fragments vaulted like rebounding hail,',
    )) == (
        'A mighty fountain momently was forced:, '
        'Amid whose swift half-intermitted burst, or '
        'Huge fragments vaulted like rebounding hail,'
    )
